from types import SimpleNamespace
from unittest.mock import MagicMock

import pytest

from refinitiv.data.content import trade_data_service
from tests.unit.conftest import StubSession
import refinitiv.data as rd
from refinitiv.data._core.session import GrantPassword, EventCode
from refinitiv.data.session import platform, desktop


@pytest.mark.parametrize(
    "defn_with_stream",
    [
        rd.content.pricing.Definition("universe"),
        rd.delivery.rdp_stream.Definition("", [], [], {}, "test.api"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.custom_instruments.Definition("universe"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.ipa.financial_contracts.cap_floor.Definition("universe"),
        rd.content.ipa.financial_contracts.cds.Definition("universe"),
        rd.content.ipa.financial_contracts.cross.Definition("universe"),
        rd.content.ipa.financial_contracts.swap.Definition("universe"),
        rd.content.ipa.financial_contracts.option.Definition(),
        rd.content.ipa.financial_contracts.repo.Definition(),
        rd.content.ipa.financial_contracts.swaption.Definition(),
        rd.content.ipa.financial_contracts.term_deposit.Definition(),
    ],
    ids=[
        "pricing",
        "rdp_stream",
        "omm_stream",
        "custom_instruments",
        "financial_contracts.bond",
        "financial_contracts.cap_floor",
        "financial_contracts.cds",
        "financial_contracts.cross",
        "financial_contracts.swap",
        "financial_contracts.option",
        "financial_contracts.repo",
        "financial_contracts.swaption",
        "financial_contracts.term_deposit",
    ],
)
def test_on_update(defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_update_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_update(on_update_callback)
    stream._stream._on_stream_update(MagicMock(), {"Fields": {}})

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "defn_with_stream",
    [
        rd.content.pricing.Definition("universe"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=[
        "pricing",
        "omm_stream",
        "custom_instruments",
    ],
)
def test_on_refresh(defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_refresh_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_refresh(on_refresh_callback)
    stream._stream._on_stream_refresh(MagicMock(), {"Fields": {}})

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "defn_with_stream",
    [
        rd.content.pricing.Definition("universe"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=[
        "pricing",
        "omm_stream",
        "custom_instruments",
    ],
)
def test_on_status(defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_status_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_status(on_status_callback)
    stream._stream._on_stream_status(MagicMock(), {"Fields": {}})

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "mock_name, obj_definition",
    [
        ("universe", rd.content.pricing.Definition("universe")),
        ("name", rd.delivery.omm_stream.Definition("name")),
        ("S)name.", rd.content.custom_instruments.Definition("name")),
    ],
    ids=[
        "pricing",
        "omm_stream",
        "custom_instruments",
    ],
)
def test_on_complete(mock_name, obj_definition):
    # given
    session = StubSession(is_open=True)
    stream = obj_definition.get_stream(session)

    mock_stream = SimpleNamespace()
    mock_stream.name = mock_name

    callback_called = False

    def on_complete_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_complete(on_complete_callback)
    stream._stream._on_stream_complete(mock_stream)

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "defn_with_stream",
    [
        rd.content.pricing.Definition("universe"),
        rd.delivery.omm_stream.Definition(""),
        rd.content.custom_instruments.Definition("universe"),
    ],
    ids=[
        "pricing",
        "omm_stream",
        "custom_instruments",
    ],
)
def test_on_error(defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_error_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_error(on_error_callback)
    stream._stream._on_stream_error(MagicMock(), {"Fields": {}})

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "defn_with_stream",
    [
        rd.delivery.rdp_stream.Definition("", [], [], {}, "test.api"),
        rd.delivery.omm_stream.Definition(""),
    ],
    ids=[
        "rdp_stream",
        "omm_stream",
    ],
)
def test_on_ack(defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_ack_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_ack(on_ack_callback)
    stream._stream._on_stream_ack(MagicMock(), {"Fields": {}})

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "defn_with_stream",
    [
        rd.delivery.rdp_stream.Definition("", [], [], {}, "test.api"),
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.ipa.financial_contracts.cap_floor.Definition("universe"),
        rd.content.ipa.financial_contracts.cds.Definition("universe"),
        rd.content.ipa.financial_contracts.cross.Definition("universe"),
        rd.content.ipa.financial_contracts.swap.Definition("universe"),
        rd.content.ipa.financial_contracts.option.Definition(),
        rd.content.ipa.financial_contracts.repo.Definition(),
        rd.content.ipa.financial_contracts.swaption.Definition(),
        rd.content.ipa.financial_contracts.term_deposit.Definition(),
    ],
    ids=[
        "rdp_stream",
        "financial_contracts.bond",
        "financial_contracts.cap_floor",
        "financial_contracts.cds",
        "financial_contracts.cross",
        "financial_contracts.swap",
        "financial_contracts.option",
        "financial_contracts.repo",
        "financial_contracts.swaption",
        "financial_contracts.term_deposit",
    ],
)
def test_on_response(defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_response_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_response(on_response_callback)
    stream._stream._on_stream_response(MagicMock(), {"Fields": {}})

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "defn_with_stream",
    [
        rd.delivery.rdp_stream.Definition("", [], [], {}, "test.api"),
    ],
    ids=[
        "rdp_stream",
    ],
)
def test_on_alarm(defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_alarm_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_alarm(on_alarm_callback)
    stream._stream._on_stream_alarm(MagicMock(), {"Fields": {}})

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "defn_with_stream",
    [
        rd.content.ipa.financial_contracts.bond.Definition("universe"),
        rd.content.ipa.financial_contracts.cap_floor.Definition("universe"),
        rd.content.ipa.financial_contracts.cds.Definition("universe"),
        rd.content.ipa.financial_contracts.cross.Definition("universe"),
        rd.content.ipa.financial_contracts.swap.Definition("universe"),
        rd.content.ipa.financial_contracts.option.Definition(),
        rd.content.ipa.financial_contracts.repo.Definition(),
        rd.content.ipa.financial_contracts.swaption.Definition(),
        rd.content.ipa.financial_contracts.term_deposit.Definition(),
    ],
    ids=[
        "financial_contracts.bond",
        "financial_contracts.cap_floor",
        "financial_contracts.cds",
        "financial_contracts.cross",
        "financial_contracts.swap",
        "financial_contracts.option",
        "financial_contracts.repo",
        "financial_contracts.swaption",
        "financial_contracts.term_deposit",
    ],
)
def test_on_state(defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_state_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_state(on_state_callback)
    stream._stream._on_stream_alarm(MagicMock(), {"state": {}})

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "obj_session",
    [
        desktop.Definition(app_key="foo"),
        platform.Definition(
            app_key="foo", grant=GrantPassword(username="username", password="password")
        ),
    ],
    ids=[
        "session.desktop",
        "session.platform",
    ],
)
def test_on_state_for_session(obj_session):
    # given
    session = obj_session.get_session()

    callback_called = False

    def on_state_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    session.on_state(on_state_callback)
    session._call_on_state("")

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "obj_session",
    [
        desktop.Definition(app_key="foo"),
        platform.Definition(
            app_key="foo", grant=GrantPassword(username="username", password="password")
        ),
    ],
    ids=[
        "session.desktop",
        "session.platform",
    ],
)
def test_on_event_for_session(obj_session):
    # given
    session = obj_session.get_session()

    callback_called = False

    def on_event_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    session.on_event(on_event_callback)
    session._call_on_event(EventCode.SessionConnected, "")

    # then
    assert callback_called is True


@pytest.mark.parametrize(
    "callback, input_message",
    [
        ("on_update", {"update": [{}]}),
        ("on_complete", {"state": {"message": "queueSize=0"}}),
        ("on_add", {"data": [{}]}),
        ("on_remove", {"remove": [{}]}),
        ("on_event", {"messages": [{}]}),
        ("on_state", {"state": {"test": None}}),
    ],
    ids=[
        "trade_data_service.on_update",
        "trade_data_service.on_complete",
        "trade_data_service.on_add",
        "trade_data_service.on_remove",
        "trade_data_service.on_event",
        "trade_data_service.on_state",
    ],
)
def test_callbacks_for_trade_data_service(callback, input_message):
    # given
    session = StubSession(is_open=True)
    defn_with_stream = trade_data_service.Definition()
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def _callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    getattr(stream, callback)(_callback)
    stream._stream._do_on_update(MagicMock(), input_message)

    # then
    assert callback_called is True


def test_on_update_for_price_chain():
    # given
    session = StubSession(is_open=True)
    defn_with_stream = rd.content.pricing.chain.Definition("universe")
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_update(on_callback)
    stream._stream.dispatch_update(...)

    # then
    assert callback_called is True


def test_on_add_for_price_chain():
    # given
    session = StubSession(is_open=True)
    defn_with_stream = rd.content.pricing.chain.Definition("universe")
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_add(on_callback)
    stream._stream.dispatch_add(..., ...)

    # then
    assert callback_called is True


def test_on_complete_for_price_chain():
    # given
    session = StubSession(is_open=True)
    defn_with_stream = rd.content.pricing.chain.Definition("universe")
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_complete(on_callback)
    stream._stream.dispatch_complete(...)

    # then
    assert callback_called is True


def test_on_remove_for_price_chain():
    # given
    session = StubSession(is_open=True)
    defn_with_stream = rd.content.pricing.chain.Definition("universe")
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_remove(on_callback)
    stream._stream.dispatch_remove(..., ...)

    # then
    assert callback_called is True


def test_on_error_for_price_chain():
    # given
    session = StubSession(is_open=True)
    defn_with_stream = rd.content.pricing.chain.Definition("universe")
    stream = defn_with_stream.get_stream(session)

    callback_called = False

    def on_callback(*args, **kwargs):
        nonlocal callback_called
        callback_called = True

    # when
    stream.on_error(on_callback)
    stream._stream.dispatch_error(...)

    # then
    assert callback_called is True



# coding: utf8

from typing import Optional

from ._enums import PriceSide
from ..._object_definition import ObjectDefinition


class PricingParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    price_side : PriceSide, optional
        Price Side to consider when retrieving Market Data.
    income_tax_percent : float, optional
        Income tax percent is subtracted from applied interest rate percents at the end
        of the deposit. Example: "5" means 5%
        By default is 0.
    market_data_date : str, optional
        The market data date for pricing.
        By default, the market_data_date date is the valuation_date or Today.
    report_ccy : str, optional
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). It is set for the fields ending with 'xxxinreportccy'. Optional. The
        default value is the notional currency.
    valuation_date : str, optional
        The valuation date for pricing. If not set the valuation date is equal to
        market_data_date or Today. For assets that contains a settlementConvention, the
        default valuation date  is equal to the settlementdate of the Asset that is
        usually the TradeDate+SettlementConvention.

    Examples
    --------
     >>> import refinitiv.data.content.ipa.financial_contracts as rdf
     >>> rdf.term_deposit.PricingParameters(valuation_date="2020-04-24")
    """

    def __init__(
        self,
        price_side: Optional[PriceSide] = None,
        income_tax_percent: Optional[float] = None,
        market_data_date: Optional[str] = None,
        report_ccy: Optional[str] = None,
        valuation_date: Optional[str] = None,
    ):
        super().__init__()
        self.price_side = price_side
        self.income_tax_percent = income_tax_percent
        self.market_data_date = market_data_date
        self.report_ccy = report_ccy
        self.valuation_date = valuation_date

    @property
    def price_side(self):
        """
        Price Side to consider when retrieving Market Data.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def income_tax_percent(self):
        """
        Income tax percent is substracted from applied interest rate percents in the end of deposit.
        Example: "5" means 5%
        :return: float
        """
        return self._get_parameter("incomeTaxPercent")

    @income_tax_percent.setter
    def income_tax_percent(self, value):
        self._set_parameter("incomeTaxPercent", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing.
        By default, the marketDataDate date is the ValuationDate or Today.
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def report_ccy(self):
        """
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def valuation_date(self):
        """
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



import urllib
from typing import TYPE_CHECKING

from refinitiv.data._tools import urljoin
from ._data_types import BucketData, FileData, FileSetData, PackageData
from ._iter_object import IterObj
from ._tools import _get_query_params
from .._data._data_provider import (
    DataProvider,
    RequestFactory,
)
from .._data._response_factory import ResponseFactory

if TYPE_CHECKING:
    from .._data._data_provider import ParsedData
    from ..._core.session import Session


# --------------------------------------------------------------------------------------
#   Request factory
# --------------------------------------------------------------------------------------


class CFSRequestFactory(RequestFactory):
    def get_query_parameters(self, *_, **kwargs) -> list:
        return _get_query_params(**kwargs)

    def add_query_parameters(self, url, query_parameters) -> str:
        return "?".join([url, urllib.parse.urlencode(query_parameters, safe=";")])


class CFSPackageRequestFactory(CFSRequestFactory):
    def get_query_parameters(self, *_, **kwargs) -> list:
        package_id = kwargs.get("_package_id")
        if package_id is not None:
            return []

        return super().get_query_parameters(**kwargs)

    def get_url(self, *args, **kwargs):
        url = super().get_url(*args, **kwargs)
        url_id = kwargs.get("_package_id")
        if url_id is not None:
            url = urljoin(url, url_id)
        return url


class CFSStreamRequestFactory(RequestFactory):
    def get_url(self, *args, **kwargs):
        return super().get_url(*args, **kwargs) + "/{id}/stream"

    def get_path_parameters(
        self, session=None, *, path_parameters=None, id=None, **kwargs
    ) -> dict:
        path_parameters = path_parameters or {}
        if id:
            path_parameters["id"] = id
        return path_parameters

    def get_query_parameters(self, *_, **kwargs) -> list:
        query_parameters = kwargs.get("query_parameters") or []
        query_parameters.append(("doNotRedirect", "true"))
        return query_parameters


# --------------------------------------------------------------------------------------
#   Response factory
# --------------------------------------------------------------------------------------


class CFSResponseFactory(ResponseFactory):
    def create_data_success(
        self, parsed_data: "ParsedData", session: "Session" = None, **kwargs
    ):
        raw = self.get_raw(parsed_data)
        content_value = raw.get("value") or [raw]
        return self.data_class(raw, IterObj(content_value, session, self.data_class))


# --------------------------------------------------------------------------------------
#   Data provider
# --------------------------------------------------------------------------------------


cfs_request_factory = CFSRequestFactory()
cfs_buckets_data_provider = DataProvider(
    request=cfs_request_factory, response=CFSResponseFactory(data_class=BucketData)
)
cfs_file_sets_data_provider = DataProvider(
    request=cfs_request_factory, response=CFSResponseFactory(data_class=FileSetData)
)
cfs_files_data_provider = DataProvider(
    request=cfs_request_factory, response=CFSResponseFactory(data_class=FileData)
)
cfs_packages_data_provider = DataProvider(
    request=CFSPackageRequestFactory(),
    response=CFSResponseFactory(data_class=PackageData),
)
cfs_stream_data_provider = DataProvider(request=CFSStreamRequestFactory())

del cfs_request_factory



import pytest

from refinitiv.data._core.session._session_cxn_type import SessionCxnType
from refinitiv.data.delivery._stream._stream_cxn_config_provider import (
    PlatformCxnConfigProvider,
    get_discovery_url,
)
from tests.unit.conftest import StubSession, StubConfig, StubResponse, args


def create_platform_cxn_config_provider(rawconfig):
    from tests.unit.delivery.stream.raw_infos import data

    response = StubResponse(data)
    config = StubConfig(rawconfig)
    session = StubSession(
        config=config, response=response, session_cxn_type=SessionCxnType.REFINITIV_DATA
    )
    provider = PlatformCxnConfigProvider()
    return provider, session


@pytest.mark.parametrize(
    "rawconfig,expected_url",
    [
        args(
            input={
                "apis.streaming.middle.url": "/head/middle/v1",
                "apis.streaming.middle.endpoints.tail.path": "/",
                "apis.streaming.middle.endpoints.tail.locations": [],
                "apis.streaming.middle.endpoints.tail.protocols": [],
            },
            expected="wss://ap-southeast-1-aws-1-lrg.optimized-pricing:443/api.refinitiv.net",
        ),
        args(
            input={
                "apis.streaming.middle.url": "/head/middle/v1",
                "apis.streaming.middle.endpoints.tail.path": "/",
                "apis.streaming.middle.endpoints.tail.locations": "amer",
                "apis.streaming.middle.endpoints.tail.protocols": "OMM",
            },
            expected="wss://ap-southeast-1-aws-1-lrg.optimized-pricing:443/api.refinitiv.net",
        ),
        args(
            input={
                "apis.streaming.middle.endpoints.tail.direct-url": "wss://amer-3-t3.beta.streaming-pricing-api.refinitiv.com:443",
                "apis.streaming.middle.endpoints.tail.protocols": [],
            },
            expected="wss://amer-3-t3.beta.streaming-pricing-api.refinitiv.com:443/WebSocket",
        ),
        args(
            input={
                "apis.streaming.middle.endpoints.tail.direct-url": "//amer-3-t3.beta.streaming-pricing-api.refinitiv.com:443",
                "apis.streaming.middle.endpoints.tail.protocols": [],
            },
            expected="wss://amer-3-t3.beta.streaming-pricing-api.refinitiv.com:443/WebSocket",
        ),
        args(
            input={
                "apis.streaming.middle.endpoints.tail.direct-url": "//amer-3-t3.beta.streaming-pricing-api.refinitiv.com",
                "apis.streaming.middle.endpoints.tail.protocols": [],
            },
            expected="ws://amer-3-t3.beta.streaming-pricing-api.refinitiv.com:80/WebSocket",
        ),
        args(
            input={
                "apis.streaming.middle.endpoints.tail.direct-url": "amer-3-t3.beta.streaming-pricing-api.refinitiv.com",
                "apis.streaming.middle.endpoints.tail.protocols": [],
            },
            expected="ws://amer-3-t3.beta.streaming-pricing-api.refinitiv.com:80/WebSocket",
        ),
    ],
)
def test_configuration_parsed_correct(rawconfig, expected_url):
    # given
    provider, session = create_platform_cxn_config_provider(rawconfig)

    # when
    config = provider.get_cfg(session, "apis.streaming.middle.endpoints.tail")

    # then
    assert config.url == expected_url


def test_will_raise_error_if_no_url_in_config():
    # given
    provider, session = create_platform_cxn_config_provider({})

    # then
    with pytest.raises(KeyError, match="apis.streaming.middle.url"):
        # when
        provider.get_cfg(session, "apis.streaming.middle.endpoints.tail")


def test_will_raise_error_if_no_path_in_config():
    # given
    provider, session = create_platform_cxn_config_provider(
        {
            "apis.streaming.middle.url": "/head/middle/v1",
        }
    )

    # then
    with pytest.raises(KeyError, match="Cannot find discovery endpoint"):
        # when
        provider.get_cfg(session, "apis.streaming.middle.endpoints.tail")


def test_will_raise_error_if_no_locations_in_config():
    # given
    provider, session = create_platform_cxn_config_provider(
        {
            "apis.streaming.middle.url": "/head/middle/v1",
            "apis.streaming.middle.endpoints.tail.path": "/",
        }
    )

    # then
    with pytest.raises(
        KeyError, match="apis.streaming.middle.endpoints.tail.locations"
    ):
        # when
        provider.get_cfg(session, "apis.streaming.middle.endpoints.tail")


def test_will_raise_error_if_no_protocols_in_config():
    # given
    provider, session = create_platform_cxn_config_provider(
        {
            "apis.streaming.middle.url": "/head/middle/v1",
            "apis.streaming.middle.endpoints.tail.path": "/",
            "apis.streaming.middle.endpoints.tail.locations": [],
        }
    )

    # then
    with pytest.raises(
        KeyError, match="apis.streaming.middle.endpoints.tail.protocols"
    ):
        # when
        provider.get_cfg(session, "apis.streaming.middle.endpoints.tail")


def test_infos_parsed_correct():
    # given
    provider, session = create_platform_cxn_config_provider(
        {
            "apis.streaming.middle.url": "http://example.dom/head/middle/v1",
            "apis.streaming.middle.endpoints.tail.path": "/",
            "apis.streaming.middle.endpoints.tail.locations": [],
            "apis.streaming.middle.endpoints.tail.protocols": [],
        }
    )

    # when
    config = provider.get_cfg(session, "apis.streaming.middle.endpoints.tail")

    # then
    assert len(config.urls) == 12


def test_infos_filtered_by_location_correct():
    # given
    expected_location = "ap-southeast-1a"
    provider, session = create_platform_cxn_config_provider(
        {
            "apis.streaming.middle.url": "path/to/head/middle/v1",
            "apis.streaming.middle.endpoints.tail.path": "/",
            "apis.streaming.middle.endpoints.tail.locations": [expected_location],
            "apis.streaming.middle.endpoints.tail.protocols": [],
        }
    )

    # when
    config = provider.get_cfg(session, "apis.streaming.middle.endpoints.tail")

    # then
    assert len(config.urls) == 2


def test_infos_filtered_by_location_no_duplication():
    # given
    expected_locations = ["ap-southeast-1a", "ap-southeast-1b"]
    provider, session = create_platform_cxn_config_provider(
        {
            "apis.streaming.middle.url": "path/to/head/middle/v1",
            "apis.streaming.middle.endpoints.tail.path": "/",
            "apis.streaming.middle.endpoints.tail.locations": expected_locations,
            "apis.streaming.middle.endpoints.tail.protocols": [],
        }
    )

    # when
    config = provider.get_cfg(session, "apis.streaming.middle.endpoints.tail")

    # then
    assert len(config.urls) == 3


@pytest.mark.parametrize(
    "config,expected_url",
    [
        args(
            config=StubConfig(
                {
                    "apis.streaming.streaming_name.url": "/streaming_name/path/v1",
                    "apis.streaming.streaming_name.endpoints.endpoint_name.path": "/",
                }
            ),
            expected="https://root_url/streaming_name/path/v1/",
        ),
        args(
            config=StubConfig(
                {
                    "apis.streaming.streaming_name.url": "https://path.com/streaming_name/path/v1",
                    "apis.streaming.streaming_name.endpoints.endpoint_name.path": "/",
                }
            ),
            expected="https://path.com/streaming_name/path/v1/",
        ),
        args(
            config=StubConfig(
                {
                    "apis.streaming.streaming_name.url": "://path.com/streaming_name/path/v1",
                    "apis.streaming.streaming_name.endpoints.endpoint_name.path": "/",
                }
            ),
            expected="https://root_url/://path.com/streaming_name/path/v1/",
        ),
        args(
            config=StubConfig(
                {
                    "apis.streaming.streaming_name.url": "//path.com/streaming_name/path/v1",
                    "apis.streaming.streaming_name.endpoints.endpoint_name.path": "/",
                }
            ),
            expected="https://root_url/path.com/streaming_name/path/v1/",
        ),
        args(
            config=StubConfig(
                {
                    "apis.streaming.streaming_name.url": "/path.com/streaming_name/path/v1",
                    "apis.streaming.streaming_name.endpoints.endpoint_name.path": "/",
                }
            ),
            expected="https://root_url/path.com/streaming_name/path/v1/",
        ),
    ],
)
def test_get_discovery_url(config, expected_url):
    # when
    base_streaming_name = "apis.streaming.streaming_name"
    full_streaming_name = f"{base_streaming_name}.endpoints.endpoint_name"
    root_url = "https://root_url"
    testing_url = get_discovery_url(
        root_url, base_streaming_name, full_streaming_name, config
    )

    # then
    assert testing_url == expected_url, testing_url


def test_tcp_filtering():
    provider, session = create_platform_cxn_config_provider(
        {
            "apis.streaming.pricing.url": "http://example.dom/apis/pricing/v5",
            "apis.streaming.pricing.endpoints.endpoint.path": "/",
            "apis.streaming.pricing.endpoints.endpoint.locations": [],
            "apis.streaming.pricing.endpoints.endpoint.protocols": [],
            "apis.streaming.pricing.use_rwf": True,
        }
    )

    # when
    config = provider.get_cfg(session, "apis.streaming.pricing.endpoints.endpoint")

    # then
    assert len(config.urls) == 12
    assert all(url.endswith(":14002") for url in config.urls)


def test_tier_filtering():
    # given
    provider, session = create_platform_cxn_config_provider(
        {
            "apis.streaming.pricing.url": "http://example.dom/apis/pricing/v5",
            "apis.streaming.pricing.endpoints.endpoint.path": "/",
            "apis.streaming.pricing.endpoints.endpoint.locations": [],
            "apis.streaming.pricing.endpoints.endpoint.protocols": [],
            "apis.streaming.pricing.use_rwf": True,
            "apis.streaming.pricing.tier": 1500,
        }
    )

    # when
    config = provider.get_cfg(session, "apis.streaming.pricing.endpoints.endpoint")

    # then
    assert len(config.urls) == 11



# coding: utf8

from ..._object_definition import ObjectDefinition


class Instrument(ObjectDefinition):
    def __init__(self, instrument_code=None, value=None):
        super().__init__()
        self.instrument_code = instrument_code
        self.value = value

    @property
    def instrument_code(self):
        """
        :return: str
        """
        return self._get_parameter("instrumentCode")

    @instrument_code.setter
    def instrument_code(self, value):
        self._set_parameter("instrumentCode", value)

    @property
    def value(self):
        """
        :return: float
        """
        return self._get_parameter("value")

    @value.setter
    def value(self, value):
        self._set_parameter("value", value)



from types import SimpleNamespace

import pytest
from pandas.core.dtypes.common import is_datetime64_any_dtype, is_datetime64tz_dtype


class ExceptionInjection:
    _stream = SimpleNamespace()
    _stream.universe = None

    def __init__(self, *args, **kwargs):
        raise ValueError


args_names = (
    "adc_df,pricing_df,data,expected_result,adc_stub,pricing_stub,exception_expected"
)


def assert_datetime_dtype_for_df(df):
    dict_values = df.dtypes.to_dict().values()
    assert any(is_datetime64_any_dtype(dtype) for dtype in dict_values)

    for dtype in dict_values:
        if is_datetime64_any_dtype(dtype):
            assert not is_datetime64tz_dtype(dtype)



import pytest
import refinitiv.data as rd
from types import ModuleType

testing_modules = [
    rd,
    rd.content,
    rd.content.custom_instruments,
    rd.content.custom_instruments.events,
    rd.content.custom_instruments.manage,
    rd.content.custom_instruments.search,
    rd.content.custom_instruments.summaries,
    rd.dates_and_calendars,
    rd.news,
    rd.content.esg,
    rd.content.esg.basic_overview,
    rd.content.esg.bulk,
    rd.content.esg.full_measures,
    rd.content.esg.full_scores,
    rd.content.esg.standard_measures,
    rd.content.esg.standard_scores,
    rd.content.esg.universe,
    rd.content.estimates,
    rd.content.estimates.view_actuals,
    rd.content.estimates.view_actuals.annual,
    rd.content.estimates.view_actuals.interim,
    rd.content.estimates.view_actuals_kpi,
    rd.content.estimates.view_actuals_kpi.annual,
    rd.content.estimates.view_actuals_kpi.interim,
    rd.content.estimates.view_summary,
    rd.content.estimates.view_summary.annual,
    rd.content.estimates.view_summary.historical_snapshots_non_periodic_measures,
    rd.content.estimates.view_summary.historical_snapshots_periodic_measures_annual,
    rd.content.estimates.view_summary.historical_snapshots_periodic_measures_interim,
    rd.content.estimates.view_summary.historical_snapshots_recommendations,
    rd.content.estimates.view_summary.interim,
    rd.content.estimates.view_summary.non_periodic_measures,
    rd.content.estimates.view_summary.recommendations,
    rd.content.estimates.view_summary_kpi,
    rd.content.estimates.view_summary_kpi.annual,
    rd.content.estimates.view_summary_kpi.historical_snapshots_kpi,
    rd.content.estimates.view_summary_kpi.interim,
    rd.content.filings,
    rd.content.filings.retrieval,
    rd.content.filings.search,
    rd.content.fundamental_and_reference,
    rd.content.historical_pricing,
    rd.content.historical_pricing.events,
    rd.content.historical_pricing.summaries,
    rd.content.ipa,
    rd.content.ipa.curves,
    rd.content.ipa.curves.forward_curves,
    rd.content.ipa.curves.zc_curve_definitions,
    rd.content.ipa.curves.zc_curves,
    rd.content.ipa.dates_and_calendars,
    rd.content.ipa.dates_and_calendars.add_periods,
    rd.content.ipa.dates_and_calendars.count_periods,
    rd.content.ipa.dates_and_calendars.date_schedule,
    rd.content.ipa.dates_and_calendars.holidays,
    rd.content.ipa.dates_and_calendars.is_working_day,
    rd.content.ipa.financial_contracts,
    rd.content.ipa.financial_contracts.bond,
    rd.content.ipa.financial_contracts.cap_floor,
    rd.content.ipa.financial_contracts.cds,
    rd.content.ipa.financial_contracts.cross,
    rd.content.ipa.financial_contracts.option,
    rd.content.ipa.financial_contracts.repo,
    rd.content.ipa.financial_contracts.swap,
    rd.content.ipa.financial_contracts.swaption,
    rd.content.ipa.financial_contracts.term_deposit,
    rd.content.ipa.surfaces,
    rd.content.ipa.surfaces.cap,
    rd.content.ipa.surfaces.eti,
    rd.content.ipa.surfaces.fx,
    rd.content.ipa.surfaces.swaption,
    rd.content.news,
    rd.content.news.headlines,
    rd.content.news.story,
    rd.content.news.top_news,
    rd.content.news.top_news.hierarchy,
    rd.content.news.images,
    rd.content.ownership,
    rd.content.ownership.consolidated,
    rd.content.ownership.consolidated.breakdown,
    rd.content.ownership.consolidated.concentration,
    rd.content.ownership.consolidated.investors,
    rd.content.ownership.consolidated.recent_activity,
    rd.content.ownership.consolidated.shareholders_history_report,
    rd.content.ownership.consolidated.shareholders_report,
    rd.content.ownership.consolidated.top_n_concentration,
    rd.content.ownership.fund,
    rd.content.ownership.fund.breakdown,
    rd.content.ownership.fund.concentration,
    rd.content.ownership.fund.holdings,
    rd.content.ownership.fund.investors,
    rd.content.ownership.fund.recent_activity,
    rd.content.ownership.fund.shareholders_history_report,
    rd.content.ownership.fund.shareholders_report,
    rd.content.ownership.fund.top_n_concentration,
    rd.content.ownership.insider,
    rd.content.ownership.insider.shareholders_report,
    rd.content.ownership.insider.transaction_report,
    rd.content.ownership.investor,
    rd.content.ownership.investor.holdings,
    rd.content.ownership.org_info,
    rd.content.pricing,
    rd.content.pricing.chain,
    rd.content.search,
    rd.content.search.lookup,
    rd.content.search.metadata,
    rd.content.symbol_conversion,
    rd.content.trade_data_service,
    rd.delivery,
    rd.delivery.cfs,
    rd.delivery.cfs.buckets,
    rd.delivery.cfs.file_downloader,
    rd.delivery.cfs.file_sets,
    rd.delivery.cfs.files,
    rd.delivery.cfs.packages,
    rd.delivery.endpoint_request,
    rd.delivery.omm_stream,
    rd.delivery.rdp_stream,
    rd.eikon,
    rd.errors,
    rd.discovery,
    rd.session,
    rd.session.desktop,
    rd.session.platform,
    rd.usage_collection,
]


@pytest.mark.parametrize("module", testing_modules)
def test_import_all(module):
    """Test that __all__ contains only names that are actually exported."""
    missing = set(n for n in module.__all__ if getattr(module, n, None) is None)
    assert not missing, f"__all__ contains unresolved names: {', '.join(missing)}"


def test_number_modules():
    counter = 0

    def count_module(pack):
        nonlocal counter
        for s in dir(pack):
            if s.startswith("_"):
                continue

            attr = getattr(pack, s)

            if isinstance(attr, ModuleType):
                counter += 1
                count_module(attr)

    # given
    expected_number = 124

    # when
    count_module(rd)
    existing_number = counter + 1  # 1 for rd
    testing_number = len(testing_modules)

    # then
    assert existing_number == expected_number, existing_number
    assert testing_number == existing_number, testing_number



from typing import TYPE_CHECKING, Any

from .rdp_stream import RDPStream
from ..._tools import try_copy_to_list

if TYPE_CHECKING:
    from ..._types import ExtendedParams, StrStrings
    from ..._core.session import Session


class Definition:
    """
    This class to subscribe to streaming items of RDP streaming protocol
    that exposed by the underlying of the Refinitiv Data

    Parameters
    ----------
    service: string, optional
        name of RDP service
    universe: list
        RIC to retrieve item stream.
    view: list
        data fields to retrieve item stream
    parameters: dict
        extra parameters to retrieve item stream.
    api: string
        specific name of RDP streaming defined
        in config file. i.e. 'streaming.trading-analytics.endpoints.redi'
    extended_params: dict, optional
        Specify optional params
        Default: None

    Examples
    --------
    >>> from refinitiv.data.delivery import rdp_stream
    >>> definition = rdp_stream.Definition(
    ...     service=None,
    ...     universe=[],
    ...     view=None,
    ...     parameters={"universeType": "RIC"},
    ...     api="streaming.trading-analytics.endpoints.redi",
    ...)
    """

    def __init__(
        self,
        service: str,
        universe: Any,
        view: "StrStrings",
        parameters: dict,
        api: str,
        extended_params: "ExtendedParams" = None,
    ) -> None:
        self._service = service
        if not isinstance(universe, dict):
            universe = try_copy_to_list(universe)
        self._universe = universe
        self._view = try_copy_to_list(view)
        self._parameters = parameters
        self._api = api
        self._extended_params = extended_params

    def get_stream(self, session: "Session" = None) -> RDPStream:
        stream = RDPStream(
            session=session,
            service=self._service,
            universe=self._universe,
            view=self._view,
            parameters=self._parameters,
            api=self._api,
            extended_params=self._extended_params,
        )
        return stream



from refinitiv.data.content.ipa.curves import zc_curve_definitions
from tests.unit.conftest import StubSession, StubResponse


def test_zc_curve_definition_empty_data():
    # given
    response = StubResponse({"data": [{}]})
    session = StubSession(is_open=True, response=response)
    definition = zc_curve_definitions.Definition(source="INVALID_SOURCE")

    # when
    response = definition.get_data(session)

    # then
    try:
        df = response.data.df
    except Exception as e:
        assert False, str(e)

    assert df.empty is True
    assert response.errors == []



from typing import TYPE_CHECKING, Union

from ._asset_class import (
    create_asset_class_request_strings,
    asset_class_enum_arg_parser,
)
from ._country_code import country_code_arg_parser
from ._symbol_type import SYMBOL_TYPE_VALUES, symbol_types_arg_parser
from .._content_provider_layer import ContentProviderLayer
from ..search import Views
from ..._content_type import ContentType
from ..._tools import create_repr, Copier
from ._asset_state import asset_state_enum_arg_parser
from ._asset_state import AssetState
from ._asset_class import AssetClass

if TYPE_CHECKING:
    from ._asset_state import OptAssetState
    from ._country_code import OptCountryCode
    from ._symbol_type import OptSymbolTypes
    from ._asset_class import OptAssetClass
    from ._country_code import CountryCode
    from ._symbol_type import SymbolTypes
    from ..._types import ExtendedParams, StrStrings

DEFAULT_SCOPE = "_AllUnique"


def _prepare_filter(
    asset_state: AssetState, asset_class: Union[list, AssetClass]
) -> str:
    asset_state = asset_state or AssetState.ACTIVE

    if asset_state is AssetState.ACTIVE:
        ret_val = "AssetState eq 'AC'"
    else:
        ret_val = "(AssetState ne 'AC' and AssetState ne null)"

    if asset_class and not isinstance(asset_class, list):
        asset_class = [asset_class]

    if asset_class:
        search_all_category, rcs_asset_category = create_asset_class_request_strings(
            asset_class
        )

        if search_all_category and rcs_asset_category:
            ret_val = f"{ret_val} and ({search_all_category} or {rcs_asset_category})"
        else:
            ret_val = f"{ret_val} and ({search_all_category}{rcs_asset_category})"

    return ret_val


class Definition(ContentProviderLayer):
    """
    This class describe parameters to retrieve data for symbol conversion.

    Parameters
    ----------
    symbols: str or list of str
        Single instrument or list of instruments to convert.

    from_symbol_type: SymbolTypes, optional
        Instrument code to convert from.
        Possible values: 'CUSIP', 'ISIN', 'SEDOL', 'RIC', 'ticker', 'lipperID', 'IMO'
        Default: '_AllUnique'

    to_symbol_types: SymbolTypes, str or list of str or SymbolTypes, optional
        Instrument code to convert to.
        Possible values: 'CUSIP', 'ISIN', 'SEDOL', 'RIC', 'ticker', 'lipperID', 'IMO', 'OAPermID'
        Default: all symbol types are requested

    extended_params: dict, optional
        Other parameters can be provided if necessary

    preferred_country_code: CountryCode, optional
        Unique ISO 3166 code for country

    asset_class: AssetClass, optional
        AssetClass value to build filter parameter.

    asset_state: AssetState, optional
        AssetState value to build filter parameter.

    Examples
    --------
    >>> from refinitiv.data.content import symbol_conversion
    >>> definition = symbol_conversion.Definition(
    ...     symbols=["US5949181045", "US02079K1079"],
    ...     from_symbol_type=symbol_conversion.SymbolTypes.ISIN,
    ...     to_symbol_types=[
    ...         symbol_conversion.SymbolTypes.RIC,
    ...         symbol_conversion.SymbolTypes.OA_PERM_ID
    ...     ],
    ...     preferred_country_code=symbol_conversion.CountryCode.USA,
    ...     asset_class=[
    ...        symbol_conversion.AssetClass.COMMODITIES,
    ...        symbol_conversion.AssetClass.EQUITIES,
    ...        symbol_conversion.AssetClass.WARRANTS
    ...     ],
    ...     asset_state=symbol_conversion.AssetState.INACTIVE
    ... )
    >>> response = definition.get_data()
    """

    def __init__(
        self,
        symbols: "StrStrings",
        from_symbol_type: "SymbolTypes" = DEFAULT_SCOPE,
        to_symbol_types: "OptSymbolTypes" = SYMBOL_TYPE_VALUES,
        preferred_country_code: "OptCountryCode" = None,
        asset_class: "OptAssetClass" = None,
        asset_state: "OptAssetState" = None,
        extended_params: "ExtendedParams" = None,
    ):
        super().__init__(
            content_type=ContentType.DISCOVERY_LOOKUP,
            view=Views.SEARCH_ALL,
        )

        self.symbols = symbols
        self.from_symbol_type = from_symbol_type
        self.to_symbol_types = to_symbol_types
        self.preferred_country_code = preferred_country_code
        self.asset_class = asset_class
        self.asset_state = asset_state
        self.extended_params = extended_params

    @property
    def symbols(self) -> "StrStrings":
        return self._kwargs.get("terms")

    @symbols.setter
    def symbols(self, value: "StrStrings"):
        if value:
            value = Copier.get_list(value)
            self._kwargs["terms"] = ",".join(value)

    @property
    def from_symbol_type(self) -> "SymbolTypes":
        return self._kwargs.get("scope")

    @from_symbol_type.setter
    def from_symbol_type(self, value: "SymbolTypes"):
        scope = value or DEFAULT_SCOPE
        if value and value != DEFAULT_SCOPE:
            scope = symbol_types_arg_parser.get_str(value)

        self._kwargs["scope"] = scope

    @property
    def to_symbol_types(self) -> "OptSymbolTypes":
        return self._kwargs.get("select")

    @to_symbol_types.setter
    def to_symbol_types(self, value: "OptSymbolTypes"):
        value = value and Copier.get_list(value)
        select = ["DocumentTitle"]

        if value is SYMBOL_TYPE_VALUES:
            select.extend(value)

        elif isinstance(value, list):
            select.extend(map(symbol_types_arg_parser.get_str, value))

        elif value:
            select.append(symbol_types_arg_parser.get_str(value))

        self._kwargs["select"] = ",".join(select)

    @property
    def preferred_country_code(self) -> "OptCountryCode":
        return self._kwargs.get("boost")

    @preferred_country_code.setter
    def preferred_country_code(self, value: "OptCountryCode"):
        if value:
            value = f"RCSExchangeCountry eq '{country_code_arg_parser.get_str(value)}'"
            self._kwargs["boost"] = value

    @property
    def _filter(self) -> str:
        return self._kwargs.get("filter")

    def _update_filter(self):
        self._kwargs["filter"] = _prepare_filter(self.asset_state, self.asset_class)

    @property
    def asset_state(self) -> "OptAssetState":
        return self._kwargs.get("asset_state")

    @asset_state.setter
    def asset_state(self, value: "OptAssetState"):
        if value:
            self._kwargs["asset_state"] = asset_state_enum_arg_parser.get_enum(value)
            self._update_filter()

    @property
    def asset_class(self) -> "OptAssetClass":
        return self._kwargs.get("asset_class")

    @asset_class.setter
    def asset_class(self, value: "OptAssetClass"):
        if value:
            value = Copier.get_list(value)
            self._kwargs["asset_class"] = asset_class_enum_arg_parser.get_enum(value)
            self._update_filter()

    @property
    def extended_params(self) -> "ExtendedParams":
        return self._kwargs.get("extended_params")

    @extended_params.setter
    def extended_params(self, value: "ExtendedParams"):
        if value:
            self._kwargs["extended_params"] = value

    def __repr__(self):
        return create_repr(
            self,
            middle_path="content.symbols_convention",
            content=f"{{symbols='{self.symbols}'}}",
        )



import allure
import pytest

from refinitiv.data.content import search
from refinitiv.data.content.search import Views
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.helpers import (
    check_response_status,
    check_non_empty_response_data,
    check_extended_params_were_sent_in_request,
    get_async_response_from_definition,
    check_response_dataframe_contains_columns_names,
    check_response_dataframe_contains_rows_names,
    check_empty_dataframe_in_response,
)


@allure.suite("Content object - Search Lookup")
@allure.feature("Content object - Search Lookup")
@allure.severity(allure.severity_level.CRITICAL)
class TestLookup:
    @allure.title("Send lookup request with all valid parameters and get data")
    @pytest.mark.caseid("36107140")
    @pytest.mark.smoke
    @pytest.mark.parametrize(
        "view,scope,terms,select",
        [
            (
                "Instruments",
                "RIC",
                "MSFT.O,AAPL.O,GOOG.O,KBANK.BK,SCC.BK",
                "DocumentTitle,BusinessEntity,SEDOL,CUSIP",
            )
        ],
    )
    def test_send_lookup_request_with_all_valid_parameters(
        self, open_session, view, scope, terms, select
    ):
        response = search.lookup.Definition(
            view=view,
            scope=scope,
            terms=terms,
            select=select,
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, select.split(","))
        check_response_dataframe_contains_rows_names(response, terms.split(","))

    @allure.title("Send lookup request with invalid parameter scope")
    @pytest.mark.caseid("36107142")
    @pytest.mark.parametrize(
        "view,scope,terms,select,error_message",
        [
            (
                Views.INSTRUMENTS,
                "invalid_scope",
                "MSFT.O,AAPL.O,GOOG.O,KBANK.BK,SCC.BK",
                "DocumentTitle,BusinessEntity,SEDOL,CUSIP",
                "Invalid scope: unknown property 'invalid_scope'",
            )
        ],
    )
    def test_send_lookup_request_with_invalid_parameter_scope(
        self, open_session, view, scope, terms, select, error_message
    ):
        with pytest.raises(RDError) as error:
            search.lookup.Definition(
                view=view,
                scope=scope,
                terms=terms,
                select=select,
            ).get_data()
        assert (
            str(error.value)
            == "Error code 400 | Invalid scope: unknown property 'invalid_scope'"
        )

    @allure.title("Send lookup request with invalid parameters terms/select")
    @pytest.mark.caseid("36135835")
    @pytest.mark.parametrize(
        "view,scope,terms,select",
        [
            (
                Views.INSTRUMENTS,
                "RIC",
                "invalid_terms",
                "DocumentTitle,BusinessEntity,SEDOL,CUSIP",
            ),
            (
                search.Views.INSTRUMENTS,
                "RIC",
                "MSFT.O,AAPL.O,GOOG.O,KBANK.BK,SCC.BK",
                "invalid_select",
            ),
        ],
    )
    def test_send_lookup_request_with_invalid_parameters_terms_select(
        self, open_session, view, scope, terms, select
    ):
        response = search.lookup.Definition(
            view=view,
            scope=scope,
            terms=terms,
            select=select,
        ).get_data()

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_empty_dataframe_in_response(response)

    @allure.title("Send lookup request with invalid parameter view")
    @pytest.mark.caseid("36136025")
    @pytest.mark.parametrize(
        "view,scope,terms,select,expected_error",
        [
            (
                "invalid view - string",
                "RIC",
                "MSFT.O,AAPL.O,GOOG.O,KBANK.BK,SCC.BK",
                "DocumentTitle,BusinessEntity,SEDOL,CUSIP",
                "Error code 400 | Validation error: json.View in body should be one of "
            )
        ],
    )
    def test_send_lookup_request_with_invalid_parameter_view(
        self, open_session, view, scope, terms, select, expected_error
    ):
        with pytest.raises(RDError) as error:
            search.lookup.Definition(
                view=view,
                scope=scope,
                terms=terms,
                select=select,
            ).get_data()
        assert expected_error in str(error.value)

    @allure.title(
        "Send lookup request with all valid parameters and get data asynchronously"
    )
    @pytest.mark.caseid("36136026")
    @pytest.mark.parametrize(
        "view,scope,valid_terms,invalid_terms,select",
        [
            (
                Views.INSTRUMENTS,
                "RIC",
                "MSFT.O,AAPL.O,GOOG.O,KBANK.BK,SCC.BK",
                "INVALID_SCOPE",
                "DocumentTitle,BusinessEntity,SEDOL,CUSIP",
            )
        ],
    )
    async def test_send_lookup_request_with_all_valid_parameters_and_get_data_asynchronously(
        self, open_session_async, view, scope, valid_terms, invalid_terms, select
    ):
        full_terms = valid_terms + "," + invalid_terms
        response = await get_async_response_from_definition(
            search.lookup.Definition(
                view=view,
                scope=scope,
                terms=full_terms,
                select=select,
            )
        )

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_response_dataframe_contains_columns_names(response, select.split(","))
        check_response_dataframe_contains_rows_names(response, valid_terms.split(","))

    @allure.title("Send lookup request with extended parameters")
    @pytest.mark.caseid("36136031")
    @pytest.mark.parametrize(
        "view,scope,terms,select,extended_params",
        [
            (
                Views.INSTRUMENTS,
                "RIC",
                "MSFT.O,AAPL.O,GOOG.O,KBANK.BK,SCC.BK",
                "DocumentTitle,BusinessEntity,SEDOL,CUSIP",
                {"param1": "value1"},
            )
        ],
    )
    def test_send_lookup_request_with_extended_parameters(
        self, open_session, view, scope, terms, select, extended_params
    ):
        response = search.lookup.Definition(
            view=view,
            scope=scope,
            terms=terms,
            select=select,
            extended_params=extended_params,
        ).get_data()
        check_extended_params_were_sent_in_request(response, extended_params)
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)

    @allure.title("Send lookup request when session is not opened")
    @pytest.mark.caseid("36136071")
    @pytest.mark.parametrize(
        "view,scope,terms,select",
        [
            (
                Views.INSTRUMENTS,
                "RIC",
                "MSFT.O,AAPL.O,GOOG.O,KBANK.BK,SCC.BK",
                "DocumentTitle,BusinessEntity,SEDOL,CUSIP",
            )
        ],
    )
    def test_send_lookup_request_when_session_is_not_opened(
        self, open_session, view, scope, terms, select
    ):
        session = open_session
        session.close()
        definition = search.lookup.Definition(
            view=view,
            scope=scope,
            terms=terms,
            select=select,
        )
        with pytest.raises(
            ValueError, match="Session is not opened. Can't send any request"
        ):
            definition.get_data()



from typing import TYPE_CHECKING, Union

from .._content_data import Data
from ..._tools import create_repr
from ...delivery._data._data_provider import DataProviderLayer, BaseResponse
from ..._content_type import ContentType

if TYPE_CHECKING:
    from ._views import Views
    from ..._types import ExtendedParams


class Definition(DataProviderLayer[BaseResponse[Data]]):
    """
    This class describe parameters to retrieve data for search lookup.

    Parameters
    ----------

    view : Views
        picks a subset of the data universe to search against. see Views

    terms : str
        lists the symbols to be solved

    scope : str
        identifies the symbology which 'terms' belong to

    select : str
        specifies which properties to return for each result doc

    extended_params : dict, optional
        Other parameters can be provided if necessary

    Examples
    --------
    >>> from refinitiv.data.content import search
    >>> definition = search.lookup.Definition(
    >>>     view=search.Views.SEARCH_ALL,
    >>>     scope="RIC",
    >>>     terms="A,B,NOSUCHRIC,C,D",
    >>>     select="BusinessEntity,DocumentTitle"
    >>>)
    """

    def __init__(
        self,
        view: Union["Views", str],
        terms: str,
        scope: str,
        select: str,
        extended_params: "ExtendedParams" = None,
    ):
        self._view = view
        self._terms = terms
        self._scope = scope
        self._select = select
        self._extended_params = extended_params

        super().__init__(
            data_type=ContentType.DISCOVERY_LOOKUP,
            view=self._view,
            terms=self._terms,
            scope=self._scope,
            select=self._select,
            extended_params=self._extended_params,
        )

    def __repr__(self):
        return create_repr(
            self,
            middle_path="lookup",
            content=f"{{view='{self._view}', terms='{self._terms}', scope='{self._scope}', select='{self._select}'}}",
        )



# coding: utf8

from typing import Optional

from .._enums import DoubleBinaryType
from ..._instrument_definition import ObjectDefinition


class FxDoubleBinaryDefinition(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    double_binary_type : DoubleBinaryType, optional
        The type of a double binary option. the possible values are:
        - doublenotouch: the option expires in-the-money if the price of the underlying
          asset fails to breach either the lower trigger or the upper trigger at any
          time prior to option expiration.
    payout_amount : float, optional
        The payout amount of the option. the default value is '1,000,000'.
    payout_ccy : str, optional
        The trade currency, which is either a domestic or foreign currency. either
        payoutccy or settlementtype can be used at a time. payoutccy="foreign currency"
        is equivalent to settlementtype ="physical", and payoutccy="domestic currency"
        is equivalent to settlementtype ="cash". the value is expressed in iso 4217
        alphabetical format (e.g. 'usd').
    trigger_down : float, optional
        The lower trigger of the binary option.
    trigger_up : float, optional
        The upper trigger of the binary option.
    """

    def __init__(
        self,
        double_binary_type: Optional[DoubleBinaryType] = None,
        payout_amount: Optional[float] = None,
        payout_ccy: Optional[str] = None,
        trigger_down: Optional[float] = None,
        trigger_up: Optional[float] = None,
    ) -> None:
        super().__init__()
        self.double_binary_type = double_binary_type
        self.payout_amount = payout_amount
        self.payout_ccy = payout_ccy
        self.trigger_down = trigger_down
        self.trigger_up = trigger_up

    @property
    def double_binary_type(self):
        """
        The type of a double binary option. the possible values are:
        - doublenotouch: the option expires in-the-money if the price of the underlying
          asset fails to breach either the lower trigger or the upper trigger at any
          time prior to option expiration.
        :return: enum DoubleBinaryType
        """
        return self._get_enum_parameter(DoubleBinaryType, "doubleBinaryType")

    @double_binary_type.setter
    def double_binary_type(self, value):
        self._set_enum_parameter(DoubleBinaryType, "doubleBinaryType", value)

    @property
    def payout_amount(self):
        """
        The payout amount of the option. the default value is '1,000,000'.
        :return: float
        """
        return self._get_parameter("payoutAmount")

    @payout_amount.setter
    def payout_amount(self, value):
        self._set_parameter("payoutAmount", value)

    @property
    def payout_ccy(self):
        """
        The trade currency, which is either a domestic or foreign currency. either
        payoutccy or settlementtype can be used at a time. payoutccy="foreign currency"
        is equivalent to settlementtype ="physical", and payoutccy="domestic currency"
        is equivalent to settlementtype ="cash". the value is expressed in iso 4217
        alphabetical format (e.g. 'usd').
        :return: str
        """
        return self._get_parameter("payoutCcy")

    @payout_ccy.setter
    def payout_ccy(self, value):
        self._set_parameter("payoutCcy", value)

    @property
    def trigger_down(self):
        """
        The lower trigger of the binary option.
        :return: float
        """
        return self._get_parameter("triggerDown")

    @trigger_down.setter
    def trigger_down(self, value):
        self._set_parameter("triggerDown", value)

    @property
    def trigger_up(self):
        """
        The upper trigger of the binary option.
        :return: float
        """
        return self._get_parameter("triggerUp")

    @trigger_up.setter
    def trigger_up(self, value):
        self._set_parameter("triggerUp", value)



from typing import TYPE_CHECKING, List, Callable

from ...._tools import cached_property
from ....delivery._data._data_provider import ContentValidator

if TYPE_CHECKING:
    from ....delivery._data._data_provider import ParsedData


class ContentDataValidator(ContentValidator):
    @classmethod
    def any_valid_content_data(cls, data: "ParsedData") -> bool:
        counter = 0
        content_data = data.content_data

        if isinstance(content_data, list):
            for item in content_data:
                if item.get("error"):
                    data.error_codes.append(item["error"]["code"])
                    data.error_messages.append(item["error"]["message"])
                    counter += 1

            if counter == len(content_data):
                return False

        return True

    @cached_property
    def validators(self) -> List[Callable[["ParsedData"], bool]]:
        return [self.content_data_is_not_none, self.any_valid_content_data]



from ..conftest import args
from . import data_for_tests as dt


TEST_CASES = {
    "Simple DataFrame not valid datetime": args(
        df=dt.INPUT_DF_NOT_VALID_DATETIME,
        pattern="Date",
        utc=None,
        delete_tz=False,
        expected_dtype="datetime64[ns]",
        expected_df=dt.EXPECTED_DF_NOT_VALID_DATETIME,
    ),
    "Simple DataFrame without timezone": args(
        df=dt.INPUT_DF_WITHOUT_TIMEZONE,
        pattern="Date",
        utc=None,
        delete_tz=False,
        expected_dtype="datetime64[ns]",
        expected_df=dt.EXPECTED_DF_WITHOUT_TIMEZONE,
    ),
    "Simple DataFrame with timezone": args(
        df=dt.INPUT_DF_WITH_TIMEZONE,
        pattern="Date",
        utc=None,
        delete_tz=False,
        expected_dtype="datetime64[ns, UTC]",
        expected_df=dt.EXPECTED_DF_WITH_TIMEZONE,
    ),
    "Simple DataFrame without timezone with flag utc=True": args(
        df=dt.INPUT_DF_WITHOUT_TIMEZONE,
        pattern="Date",
        utc=True,
        delete_tz=False,
        expected_dtype="datetime64[ns, UTC]",
        expected_df=dt.EXPECTED_DF_WITH_TIMEZONE,
    ),
    "Simple DataFrame with timezone with flag utc=True": args(
        df=dt.INPUT_DF_WITH_TIMEZONE,
        pattern="Date",
        utc=True,
        delete_tz=False,
        expected_dtype="datetime64[ns, UTC]",
        expected_df=dt.EXPECTED_DF_WITH_TIMEZONE,
    ),
    "Simple DataFrame without timezone with flag utc=None, delete_tz=True": args(
        df=dt.INPUT_DF_WITHOUT_TIMEZONE,
        pattern="Date",
        utc=None,
        delete_tz=True,
        expected_dtype="datetime64[ns]",
        expected_df=dt.EXPECTED_DF_WITHOUT_TIMEZONE,
    ),
    "Simple DataFrame with timezone with flag utc=None, delete_tz=True": args(
        df=dt.INPUT_DF_WITH_TIMEZONE,
        pattern="Date",
        utc=None,
        delete_tz=True,
        expected_dtype="datetime64[ns]",
        expected_df=dt.EXPECTED_DF_WITHOUT_TIMEZONE,
    ),
    "Simple DataFrame without timezone with flag utc=True, delete_tz=True": args(
        df=dt.INPUT_DF_WITHOUT_TIMEZONE,
        pattern="Date",
        utc=True,
        delete_tz=True,
        expected_dtype="datetime64[ns]",
        expected_df=dt.EXPECTED_DF_WITHOUT_TIMEZONE,
    ),
    "Simple DataFrame with timezone with flag utc=True, delete_tz=True": args(
        df=dt.INPUT_DF_WITH_TIMEZONE,
        pattern="Date",
        utc=True,
        delete_tz=True,
        expected_dtype="datetime64[ns]",
        expected_df=dt.EXPECTED_DF_WITHOUT_TIMEZONE,
    ),
}


TEST_CASES_1 = {
    "Simple DataFrame without timezone": args(
        df=dt.INPUT_DF_WITHOUT_TIMEZONE,
        column_name="Date",
        expected_df=dt.EXPECTED_DF_WITHOUT_TIMEZONE,
    ),
    "Simple DataFrame with timezone": args(
        df=dt.INPUT_DF_WITH_TIMEZONE,
        column_name="Date",
        expected_df=dt.EXPECTED_DF_WITHOUT_TIMEZONE,
    ),
}



import pytest

from refinitiv.data.content import (
    historical_pricing as hp,
    custom_instruments,
    ownership,
    fundamental_and_reference,
    news,
)

import refinitiv.data.content.ipa.financial_contracts as rdf
from tests.unit.conftest import StubSession, StubConfig
from tests.unit.content.data_for_test_dataframe_if_data_empty import (
    HP_EVENTS,
    CUSTOM_INSTRUMENTS_EVENTS,
    OWNERSHIP_ORG_INFO,
    FUNDAMENTAL_AND_REFERENCE_UDF,
    FUNDAMENTAL_AND_REFERENCE_RDP,
    CAP_FLOOR,
    NEWS_HEADLINES_UDF,
)


@pytest.mark.parametrize(
    "definition, mock_response",
    [
        (hp.events.Definition(universe="EUR=", end="09-09-2021T13:45"), HP_EVENTS),
        (
            custom_instruments.events.Definition(
                universe="S)Batman_9ccfafd6.ertvd-111923", end="09-09-2021T13:45"
            ),
            CUSTOM_INSTRUMENTS_EVENTS,
        ),
        (ownership.org_info.Definition("TRI.N"), OWNERSHIP_ORG_INFO),
        (
            fundamental_and_reference.Definition(["IBM"], ["TR.Volume"]),
            FUNDAMENTAL_AND_REFERENCE_UDF,
        ),
        (
            news.headlines.Definition(
                "Refinitiv", date_from="20.03.1960", date_to="20.03.1961", count=3
            ),
            NEWS_HEADLINES_UDF,
        ),
    ],
)
def test_return_empty_df_if_data_empty(definition, mock_response):
    # given
    session = StubSession(is_open=True, response=mock_response)

    # when
    response = definition.get_data(session=session)

    # then
    assert response.data.df.empty


@pytest.mark.parametrize(
    "definition, config, mock_response",
    [
        (
            fundamental_and_reference.Definition(["IBM"], ["TR.Volume"]),
            StubConfig(
                {
                    "apis.data.datagrid": {
                        "url": "/data/datagrid/beta1",
                        "underlying-platform": "rdp",
                        "endpoints.standard": "/",
                    },
                    "raise_exception_on_error": True,
                }
            ),
            FUNDAMENTAL_AND_REFERENCE_RDP,
        ),
    ],
)
def test_return_empty_df_if_data_empty_for_rdp_platform(
    definition, config, mock_response
):
    # given
    session = StubSession(response=mock_response, config=config, is_open=True)

    # when
    response = definition.get_data(session=session)

    # then
    assert response.data.df.empty


def test_return_empty_df_if_data_empty_for_financial_contracts():
    # given

    session = StubSession(is_open=True, response=CAP_FLOOR)

    # when
    definition = rdf.cap_floor.Definition(
        instrument_tag="CapOnCms",
        stub_rule=rdf.cap_floor.StubRule.MATURITY,
        notional_ccy="USD",
        start_date="2018-06-15",
        end_date="2022-06-15",
        notional_amount=1000000,
        index_name="Composite",
        index_tenor="5Y",
        interest_calculation_method="Dcb_Actual_360",
        interest_payment_frequency=rdf.cap_floor.Frequency.QUARTERLY,
        buy_sell=rdf.cap_floor.BuySell.BUY,
        cap_strike_percent=1,
        pricing_parameters=rdf.cap_floor.PricingParameters(
            skip_first_cap_floorlet=False, valuation_date="2020-02-07"
        ),
    )
    response = definition.get_data(session=session)

    # then
    assert response.data.df.empty



from itertools import product
from typing import List, Tuple, Dict

import pandas as pd
from pandas.core.tools.datetimes import DatetimeScalarOrArrayConvertible

from .._types import Strings
from .._tools._dataframe import (
    convert_str_to_datetime,
    convert_dtypes,
)


def process_by_fields(timestamp_index, listoflists, fields, headers_names):
    data = []
    index = []
    for l in listoflists:
        index.append(convert_str_to_datetime(l[timestamp_index]))
        newl = []
        for field in fields:
            if field in headers_names:
                item = l[headers_names.index(field)]
                newl.append(pd.NA if item is None else item)
            else:
                newl.append(pd.NA)
        data.append(newl)
    return data, fields, index


def process_by_headers_names(
    timestamp_index, timestamp_name, listoflists, headers_names
):
    data = []
    index = []
    for l in listoflists:
        newl = []
        for item, hdr_name in zip(l, headers_names):
            if timestamp_name == hdr_name:
                index.append(convert_str_to_datetime(item))
                continue
            newl.append(pd.NA if item is None else item)
        data.append(newl)
    headers_names.pop(timestamp_index)
    return data, headers_names, index


def process_historical_raw(
    raw: dict, fields: Strings
) -> Tuple[List[list], Strings, List[DatetimeScalarOrArrayConvertible]]:
    listoflists = raw["data"]
    headers = raw["headers"]
    headers_names = [header["name"] for header in headers]

    timestamp_name = None
    if "DATE_TIME" in headers_names:
        timestamp_name = "DATE_TIME"
    elif "DATE" in headers_names:
        timestamp_name = "DATE"

    timestamp_index = headers_names.index(timestamp_name)

    if fields:
        data, columns, index = process_by_fields(
            timestamp_index, listoflists, fields, headers_names
        )

    else:
        data, columns, index = process_by_headers_names(
            timestamp_index, timestamp_name, listoflists, headers_names
        )

    return data, columns, index


def prepare_columns(
    raws, listofcolumns, bad_raws, fields, universe, items_by_date, num_fields
):
    raw_columns = None
    listofcolumns_append = listofcolumns.append
    bad_raws_append = bad_raws.append

    for instidx, raw in enumerate(raws):
        # it means error in response for custom instruments
        if not raw:
            raw = {"universe": {"ric": universe[instidx]}}
            bad_raws_append((instidx, raw))
            continue

        # it means error in response for historical pricing
        if isinstance(raw, list):
            raw = raw[0]
            bad_raws_append((instidx, raw))
            continue

        # it means in response for historical pricing events
        if isinstance(raw, dict) and not raw.get("headers"):
            raw = {"universe": {"ric": universe[instidx]}}
            bad_raws_append((instidx, raw))
            continue

        else:
            raw_datas, raw_columns, raw_index = process_historical_raw(raw, fields)

        for date, raw_data in zip(raw_index, raw_datas):
            items = items_by_date.setdefault(date, [])
            items.append((instidx, raw_data, raw_columns))

        inst_name = raw["universe"]["ric"]
        if num_fields == 1:
            processed_columns = [inst_name]

        else:
            processed_columns = list(product([inst_name], raw_columns))

        listofcolumns_append(processed_columns)

    return raw_columns


def process_bad_raws(bad_raws, listofcolumns, last_raw_columns, fields, num_fields):
    listofcolumns_insert = listofcolumns.insert

    for idx, bad_raw in bad_raws:
        raw_columns = fields or last_raw_columns or "Field"
        inst_name = bad_raw["universe"]["ric"]
        if num_fields == 1:
            processed_columns = [inst_name]

        else:
            processed_columns = list(product([inst_name], raw_columns))

        listofcolumns_insert(idx, processed_columns)


def process_data(
    data_append, index_append, date, items, num_allcolumns, num_raws, left_num_columns
):
    prev_idx = None
    counter = 0

    template = [pd.NA] * num_allcolumns
    for instidx, raw_data, raw_columns in items:
        if (counter != 0 and counter % num_raws == 0) or prev_idx == instidx:
            index_append(date)
            data_append(template)
            template = [pd.NA] * num_allcolumns
            prev_idx = instidx

        if prev_idx is None:
            prev_idx = instidx

        counter += 1

        left_idx = left_num_columns[instidx]
        right_idx = left_idx + len(raw_columns)
        for item, i in zip(raw_data, range(left_idx, right_idx)):
            template[i] = item

    index_append(date)
    data_append(template)


class HistoricalBuilder:
    def build_one(self, raw: dict, fields: Strings, axis_name: str) -> pd.DataFrame:
        if not raw["data"]:
            return pd.DataFrame()

        inst_name = raw["universe"]["ric"]
        data, columns, index = process_historical_raw(raw, fields)
        columns = pd.Index(data=columns, name=inst_name)
        index = pd.Index(data=index, name=axis_name)
        df = pd.DataFrame(data=data, columns=columns, index=index)
        df = convert_dtypes(df)
        df.sort_index(inplace=True)
        return df

    def build(
        self, raws: List[dict], universe: Strings, fields: Strings, axis_name: str
    ) -> pd.DataFrame:
        items_by_date: Dict[str, list] = {}
        listofcolumns = []
        num_raws = len(raws)
        bad_raws = []
        num_fields = len(fields)

        last_raw_columns = prepare_columns(
            raws, listofcolumns, bad_raws, fields, universe, items_by_date, num_fields
        )

        if not items_by_date:
            return pd.DataFrame()

        if bad_raws:
            process_bad_raws(
                bad_raws, listofcolumns, last_raw_columns, fields, num_fields
            )

        left_num_columns = {
            split_idx: sum([len(subcols) for subcols in listofcolumns[:split_idx]])
            for split_idx in range(num_raws)
        }

        allcolumns = [col for subcolumns in listofcolumns for col in subcolumns]

        num_allcolumns = len(allcolumns)
        data = []
        index = []
        data_append = data.append
        index_append = index.append
        for date, items in items_by_date.items():

            num_items = len(items)

            if num_items > 1:
                process_data(
                    data_append,
                    index_append,
                    date,
                    items,
                    num_allcolumns,
                    num_raws,
                    left_num_columns,
                )

            else:
                index_append(date)
                instidx, raw_data, raw_columns = items[0]
                left = [pd.NA] * left_num_columns[instidx]
                right = [pd.NA] * (num_allcolumns - len(raw_columns) - len(left))
                data_append(left + raw_data + right)

        if num_fields == 1:
            columns = pd.Index(data=allcolumns, name=fields[0])

        else:
            columns = pd.MultiIndex.from_tuples(allcolumns)

        index = pd.Index(data=index, name=axis_name)
        df = pd.DataFrame(data=data, columns=columns, index=index)
        df = convert_dtypes(df)
        df.sort_index(inplace=True)
        return df


class CustomInstsBuilder(HistoricalBuilder):
    def build_one(self, raw: dict, fields: Strings, axis_name: str) -> pd.DataFrame:
        inst_name = raw["universe"]["ric"]
        data, columns, index = process_historical_raw(raw, fields)
        if all(i is pd.NA for j in data for i in j):
            return pd.DataFrame()
        columns = pd.Index(data=columns, name=inst_name)
        index = pd.Index(data=index, name=axis_name)
        df = pd.DataFrame(data=data, columns=columns, index=index)
        df = convert_dtypes(df)
        df.sort_index(inplace=True)
        return df


historical_builder = HistoricalBuilder()
custom_insts_builder = CustomInstsBuilder()



import pytest

from refinitiv.data.content._universe_content_validator import UniverseContentValidator
from refinitiv.data.content.estimates._enums import Package
from refinitiv.data.content.estimates._data_provider import (
    EstimatesRequestFactory,
)
from refinitiv.data.delivery._data._data_provider import ParsedData


@pytest.mark.parametrize(
    ("input_value", "expected_result"),
    [
        ({"universe": "test_value"}, [("universe", "test_value")]),
        (
            {"universe": "test_value", "package": Package.BASIC},
            [("universe", "test_value"), ("package", "basic")],
        ),
        (
            {"universe": "test_value", "package": Package.STANDARD},
            [("universe", "test_value"), ("package", "standard")],
        ),
        (
            {"universe": "test_value", "package": Package.PROFESSIONAL},
            [("universe", "test_value"), ("package", "professional")],
        ),
        (
            {"universe": ["test_value1", "test_value2"]},
            [("universe", "test_value1,test_value2")],
        ),
        (
            {"universe": "test_value", "other_field": "value"},
            [("universe", "test_value")],
        ),
    ],
)
def test_estimates_request_factory_get_query_parameters(input_value, expected_result):
    # given
    request_factory = EstimatesRequestFactory()

    # when
    result = request_factory.get_query_parameters(**input_value)

    # then
    assert result == expected_result


@pytest.mark.parametrize(
    ("input_query_parameters", "input_extended_params", "expected_result"),
    [
        ([], {"universe": "test_value"}, [("universe", "test_value")]),
        (
            [("universe", "test_value")],
            {"universe": "new_value"},
            [("universe", "new_value")],
        ),
        (
            [("universe", "test_value")],
            {"universe1": "new_value"},
            [("universe", "test_value"), ("universe1", "new_value")],
        ),
    ],
)
def test_estimates_request_factory_extend_query_params(
    input_query_parameters, input_extended_params, expected_result
):
    # given
    request_factory = EstimatesRequestFactory()

    # when
    result = request_factory.extend_query_parameters(
        input_query_parameters, input_extended_params
    )

    # then
    assert result == expected_result


@pytest.mark.parametrize(
    ("input_data", "expected_result", "expected_data"),
    [
        (
            {"content_data": {"data": "data"}},
            True,
            {
                "status": {},
                "raw_response": {},
                "content_data": {"data": "data"},
                "error_codes": [],
                "error_messages": [],
            },
        ),
        (
            {"content_data": None},
            False,
            {
                "status": {},
                "raw_response": {},
                "content_data": None,
                "error_codes": [1],
                "error_messages": ["Content data is None"],
            },
        ),
        (
            {"content_data": {"error": {"code": 500, "description": "error_desc"}}},
            False,
            {
                "status": {},
                "raw_response": {},
                "content_data": {"error": {"code": 500, "description": "error_desc"}},
                "error_codes": [500],
                "error_messages": ["error_desc"],
            },
        ),
        (
            {"content_data": {"error": {"message": "error_message"}}},
            False,
            {
                "status": {},
                "raw_response": {},
                "content_data": {"error": {"message": "error_message"}},
                "error_codes": [None],
                "error_messages": ["error_message"],
            },
        ),
    ],
)
def test_estimates_content_validator(input_data, expected_result, expected_data):
    # given
    content_validator = UniverseContentValidator()

    # when
    parsed_data = ParsedData({}, {}, **input_data)
    is_valid = content_validator.validate(parsed_data)

    # then
    assert is_valid == expected_result
    assert parsed_data == expected_data



import pytest

from refinitiv.data.content.ipa._enums import Axis
from refinitiv.data.content.ipa._surfaces._models import SurfacePoint
from refinitiv.data.content.ipa._surfaces._surfaces_data_provider import (
    OneSurfaceData,
)
from tests.unit.content.ipa._surfaces.conftest import (
    one_surface_axis_params,
    one_surface_data_json,
)


@pytest.mark.parametrize(
    "value, axis",
    [
        ("24.955", Axis.X),
        ("2021-08-20", Axis.Y),
        (63.743918703207804, Axis.Z),
    ],
)
def test_get_curve_value_has_on_axis(value, axis):
    # given
    data = OneSurfaceData(one_surface_data_json, one_surface_axis_params)

    # when
    curve = data.surface.get_curve(value, axis)

    # then
    assert curve


@pytest.mark.parametrize(
    "value, axis",
    [
        ("25", Axis.X),
        ("2021-08-25", Axis.Y),
        (70, Axis.Z),
    ],
)
def test_get_curve_value_has_not_on_axis(value, axis):
    # given
    data = OneSurfaceData(one_surface_data_json, one_surface_axis_params)

    # when
    curve = data.surface.get_curve(value, axis)

    # then
    assert curve


@pytest.mark.parametrize(
    "x, y",
    [
        ("24.955", "2021-08-20"),
    ],
)
def test_get_point_x_and_y_are_on_axis(x, y):
    # given
    data = OneSurfaceData(one_surface_data_json, one_surface_axis_params)

    # when
    point = data.surface.get_point(x, y)

    # then
    assert point


@pytest.mark.parametrize(
    "x, y",
    [
        ("25", "2021-08-25"),
    ],
)
def test_get_point_x_and_y_are_not_on_axis(x, y):
    # given
    data = OneSurfaceData(one_surface_data_json, one_surface_axis_params)

    # when
    point = data.surface.get_point(x, y)

    # then
    assert point


def test_get_axis_removed():
    """get_axis removed in favor of direct x, y, z attributes access
    https://jira.refinitiv.com/browse/EAPI-2687
    """
    data = OneSurfaceData(one_surface_data_json, one_surface_axis_params)

    with pytest.raises(AttributeError):
        data.surface.get_axis()


def test_surface_point_to_string_using_str_dunder_method():
    # given
    point = SurfacePoint(0, 0, 0)

    # when
    s = str(point)

    # then
    assert s


def test_surface_point_to_string_using_repr_dunder_method():
    # given
    point = SurfacePoint(0, 0, 0)

    # when
    s = repr(point)

    # then
    assert s



from refinitiv.data.content.ipa.surfaces import eti
from tests.unit.conftest import load_json

x_axis = eti.Axis.STRIKE
y_axis = eti.Axis.DATE
one_surface_axis_params = [(x_axis, y_axis)]
one_surface_data_json = load_json(
    "./tests/unit/content/ipa/_surfaces/one_surface_data.json"
)

surfaces_axis_params = []
x_axis = eti.Axis.STRIKE
y_axis = eti.Axis.DATE
surfaces_axis_params.append((x_axis, y_axis))
x_axis = eti.Axis.DATE
y_axis = eti.Axis.STRIKE
surfaces_axis_params.append((x_axis, y_axis))
surfaces_data_json = load_json("./tests/unit/content/ipa/_surfaces/surfaces_data.json")



from typing import List

from ._adc_context import ADCContext
from ._context import RDPMixin


class ADCRDPContext(RDPMixin, ADCContext):
    def prepare_headers(self, raw: dict, *args) -> List[List[str]]:
        fields = self.fields.raw
        field_to_idx = {field.casefold(): fields.index(field) + 2 for field in fields}

        headers = [
            {"name": "instrument", "title": "Instrument"},
            {"name": "date", "title": "Date"},
        ]

        for header in raw["headers"]:
            field = f"{header['name']}.{header['title']}".casefold()
            header_name = header["name"].casefold()

            if header_name in field_to_idx and field not in field_to_idx:
                headers.insert(field_to_idx[header_name], header)
                field_to_idx.pop(header_name)

            if field in field_to_idx:
                headers.insert(field_to_idx[field], header)
                field_to_idx.pop(field)

        for field, index in field_to_idx.items():
            headers.insert(index, {"name": field.upper(), "title": field.upper()})

        return headers

    def get_cols(self, headers, fields, *args) -> List[str]:
        cols = []
        fields = [field.casefold() for field in fields]

        for header in headers:
            header_name = header["name"]
            header_title = header["title"]
            field = f"{header_name}.{header_title}".casefold()

            if header_name.casefold() == "instrument":
                continue

            elif field in fields:
                cols.append(f"{header_name}.{header_title}")

            else:
                cols.append(header_name)

        return cols

    def is_raw_headers_in_fields(self, raw, fields) -> bool:
        fields = {field.casefold() for field in fields}
        headers = set()

        for header in raw["headers"]:
            header_name = header["name"]
            if header_name not in {"instrument", "date"} and ")" not in header_name:
                header_title = header["title"]

                if " " in header_title:
                    headers.add(header_name.casefold())

                elif header_name == "RIC" and header_title == "RIC":
                    headers.add("tr.ric")

                else:
                    if header_name.split(".")[-1] == header_title:
                        headers.add(header_name.casefold())

                    else:
                        headers.add(
                            f"{header_name.casefold()}.{header_title.casefold()}"
                        )

        return headers.issubset(fields)



from enum import Enum

import pytest

from refinitiv.data.content.ipa._object_definition import ObjectDefinition
from refinitiv.data._tools._common import is_all_same_type


class MyTestEnum(Enum):
    test = "test"


class MyTestObject(ObjectDefinition):
    def __init__(self, param_name, param_value):
        super().__init__()
        self.param_name = param_name
        self.param_value = param_value

    @property
    def param_name(self):
        return self._get_parameter("param_name")

    @param_name.setter
    def param_name(self, value):
        self._set_parameter("param_name", value)

    @property
    def param_value(self):
        return self._get_parameter("param_value")

    @param_value.setter
    def param_value(self, value):
        self._set_parameter("param_value", value)

    @property
    def leg(self):
        return self._get_object_parameter(MyTestObject, "leg")

    @leg.setter
    def leg(self, value):
        self._set_object_parameter(MyTestObject, "leg", value)


def test_definition_empty():
    od = ObjectDefinition()
    assert not od._dict


def test_set_parameter():
    value = 1
    param_name = "param"
    od = ObjectDefinition()

    od._set_parameter(param_name, value)
    assert od._dict[param_name] == value

    od._set_parameter(param_name, None)
    assert param_name not in od._dict


def test_set_enum_parameter():
    param_name = "param"
    value = MyTestEnum.test.value
    od = ObjectDefinition()

    od._set_enum_parameter(MyTestEnum, param_name, MyTestEnum.test)
    assert od._dict[param_name] == value

    od._set_enum_parameter(MyTestEnum, "param_name", None)
    assert param_name in od._dict

    od._set_enum_parameter(MyTestEnum, param_name, None)
    assert param_name not in od._dict

    od._set_enum_parameter(MyTestEnum, param_name, "test")
    assert od._dict[param_name] == value

    with pytest.raises(TypeError):
        od._set_enum_parameter(MyTestEnum, param_name, "test_111")


def test_set_object_parameter():
    param_name = "param"
    value = MyTestObject("object_param", "object_value")
    od = ObjectDefinition()

    od._set_object_parameter(MyTestObject, param_name, value)
    assert param_name in od._dict
    assert isinstance(od._dict[param_name], dict)

    od._set_enum_parameter(MyTestEnum, "param_name_111", None)
    assert param_name in od._dict

    od._set_object_parameter(MyTestObject, param_name, None)
    assert param_name not in od._dict

    od._set_object_parameter(MyTestObject, param_name, "None")
    assert param_name in od._dict
    assert od._dict[param_name] == "None"


def test_set_list_parameter():
    param_name = "param"
    value = MyTestObject("object_param", "object_value")
    od = ObjectDefinition()

    with pytest.raises(TypeError):
        od._set_list_parameter(MyTestObject, param_name, [value, "value"])

    od._set_list_parameter(MyTestObject, param_name, [value])
    assert param_name in od._dict

    od._set_list_parameter(MyTestEnum, "param_name_111", None)
    assert param_name in od._dict

    od._set_list_parameter(MyTestObject, param_name, None)
    assert param_name not in od._dict

    with pytest.raises(TypeError):
        od._set_list_parameter(MyTestObject, param_name, "test_111")


def test_get_parameter():
    param_name = "param"
    value = 1
    od = ObjectDefinition()

    assert od._get_parameter(param_name) is None

    od._set_parameter(param_name, value)

    assert od._get_parameter(param_name) == value


def test_get_enum_parameter():
    value = MyTestEnum.test
    param_name = "param"
    od = ObjectDefinition()

    assert od._get_enum_parameter(MyTestEnum, param_name) is None

    od._set_enum_parameter(MyTestEnum, param_name, value)

    assert od._get_enum_parameter(MyTestEnum, param_name) == value


def test_get_object_parameter():
    param_name = "param"
    expected = MyTestObject("object_param", "object_value")
    od = ObjectDefinition()

    test_object = od._get_object_parameter(MyTestObject, param_name)
    assert test_object is None

    od._set_object_parameter(MyTestObject, param_name, expected)
    test_object = od._get_object_parameter(MyTestObject, param_name)

    assert test_object == expected
    assert isinstance(test_object, ObjectDefinition)


def test_get_list_parameter():
    param_name = "param"
    value = MyTestObject("object_param", "object_value")
    expected = [value]
    od = ObjectDefinition()

    test_object = od._get_list_parameter(MyTestObject, param_name)
    assert test_object is None

    od._set_list_parameter(MyTestObject, param_name, expected)

    test_object = od._get_list_parameter(MyTestObject, param_name)
    assert test_object == expected
    assert is_all_same_type(ObjectDefinition, test_object)


def test_get_json_empty():
    o = ObjectDefinition()
    assert not o.get_dict()


def test_get_json():
    object_param = "object_param"
    different_object_param = "different_object_param"
    myo = MyTestObject(object_param, "object_value")

    assert myo.get_dict()["param_name"] == object_param

    myo.param_name = different_object_param

    assert myo.get_dict()["param_name"] == different_object_param

    with pytest.raises(KeyError):
        _ = myo.get_dict()["leg"]


def test_get_json_complex():
    myo = MyTestObject("object_param", "object_value")
    leg = MyTestObject("a", "b")

    assert not myo.get_dict().get("leg")

    myo.leg = leg

    assert myo.get_dict()["leg"]
    assert myo.get_dict()["leg"]["param_value"] == "b"

    leg.param_value = "c"

    assert myo.get_dict()["leg"]["param_value"] == "c"


def test_from_json():
    myo = MyTestObject("object_param", "object_value")
    new_myo = myo.from_json(myo.get_dict())
    assert myo == new_myo



# coding: utf8

from typing import Optional

from .._base import BarrierDefinition
from .._enums import (
    BarrierMode,
    InOrOut,
    UpOrDown,
)


class FxBarrierDefinition(BarrierDefinition):
    """
    Parameters
    ----------
    barrier_mode : BarrierMode, optional
        Barrier Mode of the barrier option
    in_or_out : InOrOut, optional
        In/Out property of the barrier option
    up_or_down : UpOrDown, optional
        Up/Down property of the barrier option
    level : float, optional
        Barrier of the barrier option
    rebate_amount : float, optional
        Rebate of the barrier option
    window_end_date : str, optional
        Window Start date of the barrier option
    window_start_date : str, optional
        Window Start date of the barrier option
    """

    def __init__(
        self,
        barrier_mode: Optional[BarrierMode] = None,
        in_or_out: Optional[InOrOut] = None,
        up_or_down: Optional[UpOrDown] = None,
        level: Optional[float] = None,
        rebate_amount: Optional[float] = None,
        window_end_date: Optional[str] = None,
        window_start_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.barrier_mode = barrier_mode
        self.in_or_out = in_or_out
        self.up_or_down = up_or_down
        self.level = level
        self.rebate_amount = rebate_amount
        self.window_end_date = window_end_date
        self.window_start_date = window_start_date

    @property
    def barrier_mode(self):
        """
        Barrier Mode of the barrier option
        :return: enum BarrierMode
        """
        return self._get_enum_parameter(BarrierMode, "barrierMode")

    @barrier_mode.setter
    def barrier_mode(self, value):
        self._set_enum_parameter(BarrierMode, "barrierMode", value)

    @property
    def in_or_out(self):
        """
        In/Out property of the barrier option
        :return: enum InOrOut
        """
        return self._get_enum_parameter(InOrOut, "inOrOut")

    @in_or_out.setter
    def in_or_out(self, value):
        self._set_enum_parameter(InOrOut, "inOrOut", value)

    @property
    def up_or_down(self):
        """
        Up/Down property of the barrier option
        :return: enum UpOrDown
        """
        return self._get_enum_parameter(UpOrDown, "upOrDown")

    @up_or_down.setter
    def up_or_down(self, value):
        self._set_enum_parameter(UpOrDown, "upOrDown", value)

    @property
    def level(self):
        """
        Barrier of the barrier option
        :return: float
        """
        return self._get_parameter("level")

    @level.setter
    def level(self, value):
        self._set_parameter("level", value)

    @property
    def rebate_amount(self):
        """
        Rebate of the barrier option
        :return: float
        """
        return self._get_parameter("rebateAmount")

    @rebate_amount.setter
    def rebate_amount(self, value):
        self._set_parameter("rebateAmount", value)

    @property
    def window_end_date(self):
        """
        Window Start date of the barrier option
        :return: str
        """
        return self._get_parameter("windowEndDate")

    @window_end_date.setter
    def window_end_date(self, value):
        self._set_parameter("windowEndDate", value)

    @property
    def window_start_date(self):
        """
        Window Start date of the barrier option
        :return: str
        """
        return self._get_parameter("windowStartDate")

    @window_start_date.setter
    def window_start_date(self, value):
        self._set_parameter("windowStartDate", value)



import allure
import pytest

from refinitiv.data.delivery import cfs
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.delivery.cfs.conftest import (
    check_amount_of_objects_in_response_not_bigger_than,
)
from tests.integration.delivery.cfs.file_sets.conftest import (
    get_fileset_from_bucket_with_numfiles_bigger_than,
)
from tests.integration.delivery.cfs.files.conftest import (
    get_random_file_name_from_fileset,
    check_single_file_received,
    check_all_files_contain_fileset_id,
    check_file_name_in_response,
    check_files_in_response_created_after_date,
    check_files_in_response_modified_after_date,
    check_files_in_two_responses_are_not_the_same,
)
from tests.integration.helpers import (
    check_response_status,
    check_non_empty_response_data,
    get_async_response_from_definition,
)


@allure.suite("Delivery object - CFS files")
@allure.feature("Delivery object - CFS files")
@allure.severity(allure.severity_level.CRITICAL)
class TestCfsFiles:
    @allure.title(
        "Get files with valid fileset id and default amount 25 - synchronously"
    )
    @pytest.mark.caseid("36823477")
    @pytest.mark.smoke
    def test_get_files_with_valid_fileset_id_synchronously(
        self, open_platform_session, get_random_fileset
    ):
        file_set = get_random_fileset
        response = cfs.files.Definition(file_set.id).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_amount_of_objects_in_response_not_bigger_than(25, response)

    @allure.title(
        "Get files with valid fileset id and default amount 25 - asynchronously"
    )
    @pytest.mark.caseid("36823478")
    async def test_get_files_with_valid_fileset_id_asynchronously(
        self, open_platform_session_async, get_random_fileset
    ):
        file_set = get_random_fileset

        response = await get_async_response_from_definition(
            cfs.files.Definition(file_set.id)
        )
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_amount_of_objects_in_response_not_bigger_than(25, response)

    @allure.title("Get files with invalid fileset id")
    @pytest.mark.caseid("36823479")
    def test_get_files_with_invalid_fileset_id(self, open_platform_session):
        with pytest.raises(RDError, match="Error code 404 | No resource found for:"):
            cfs.files.Definition("invalid fileset id").get_data()

    @allure.title("Get files with closed session")
    @pytest.mark.caseid("36823480")
    def test_get_files_with_closed_session(self, open_platform_session):
        session = open_platform_session
        session.close()
        files_definition = cfs.files.Definition("some id")
        with pytest.raises(
            ValueError, match="Session is not opened. Can't send any request"
        ):
            files_definition.get_data()

    @allure.title("Get files with valid file_name parameter")
    @pytest.mark.caseid("36823482")
    def test_get_files_with_valid_file_name_parameter(
        self, open_platform_session, get_random_fileset
    ):
        file_set = get_random_fileset
        file_name = get_random_file_name_from_fileset(file_set)
        response = cfs.files.Definition(file_set.id, file_name=file_name).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_single_file_received(response)
        check_all_files_contain_fileset_id(response, file_set.id)
        check_file_name_in_response(response, file_name)

    @allure.title("Get files with valid created_since param")
    @pytest.mark.parametrize("created_since", ["2021-06-25T14:00:00Z"])
    @pytest.mark.caseid("36823485")
    def test_get_files_with_valid_created_since_param(
        self, open_platform_session, get_random_fileset, created_since
    ):
        file_set = get_random_fileset
        response = cfs.files.Definition(
            file_set.id, created_since=created_since
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_files_in_response_created_after_date(response, created_since)

    @allure.title("Get files with valid modified_since param")
    @pytest.mark.parametrize("modified_since", ["2021-02-25T14:00:00Z"])
    @pytest.mark.caseid("36823486")
    def test_get_files_with_valid_modified_since_param(
        self, open_platform_session, get_random_fileset, modified_since
    ):
        file_set = get_random_fileset
        response = cfs.files.Definition(
            file_set.id, modified_since=modified_since
        ).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        check_files_in_response_modified_after_date(response, modified_since)

    @allure.title("Get files with valid skip_token parameter")
    @pytest.mark.caseid("36823487")
    def test_get_files_with_valid_skip_token_param(self, open_platform_session):
        file_set = get_fileset_from_bucket_with_numfiles_bigger_than("bulk-ESG", 5)
        response1 = cfs.files.Definition(file_set.id, page_size=5).get_data()
        skip_token = response1.data.raw["skip_token"]
        response2 = cfs.files.Definition(file_set.id, skip_token=skip_token).get_data()
        check_response_status(response2, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response2)
        check_files_in_two_responses_are_not_the_same(response1, response2)

    @allure.title("Get files with valid page_size param")
    @pytest.mark.parametrize("page_size", [3])
    @pytest.mark.caseid("36823490")
    def test_get_files_with_valid_page_size_param(
        self, open_platform_session, page_size
    ):
        file_set = get_fileset_from_bucket_with_numfiles_bigger_than("bulk-ESG", 5)
        response = cfs.files.Definition(file_set.id, page_size=page_size).get_data()
        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)
        files_count = len(response.data.raw["value"])
        assert (
            files_count <= page_size
        ), f"Files count received {files_count} more than expected: {page_size}"



from ._instruments_segment import CrossCurrencyInstrumentsSegment
from ._override_bid_ask import OverrideBidAsk
from ._override_fx_forward_turn import OverrideFxForwardTurn
from ...._object_definition import ObjectDefinition


class MixinRequest(ObjectDefinition):
    @property
    def overrides(self):
        """
        :return: list OverrideBidAsk
        """
        return self._get_list_parameter(OverrideBidAsk, "overrides")

    @overrides.setter
    def overrides(self, value):
        self._set_list_parameter(OverrideBidAsk, "overrides", value)

    @property
    def segments(self):
        """
        :return: list CrossCurrencyInstrumentsSegment
        """
        return self._get_list_parameter(CrossCurrencyInstrumentsSegment, "segments")

    @segments.setter
    def segments(self, value):
        self._set_list_parameter(CrossCurrencyInstrumentsSegment, "segments", value)

    @property
    def turns(self):
        """
        :return: list OverrideFxForwardTurn
        """
        return self._get_list_parameter(OverrideFxForwardTurn, "turns")

    @turns.setter
    def turns(self, value):
        self._set_list_parameter(OverrideFxForwardTurn, "turns", value)



from json import JSONDecodeError
from typing import TYPE_CHECKING, Tuple

import requests

from ._parsed_data import ParsedData
from ..._tools import get_response_reason

if TYPE_CHECKING:
    import httpx

success_http_codes = [
    requests.codes.ok,
    requests.codes.accepted,
    requests.codes.created,
]


class Parser:
    def process_successful_response(self, raw_response: "httpx.Response") -> ParsedData:
        status = {
            "http_status_code": raw_response.status_code,
            "http_reason": get_response_reason(raw_response),
        }

        media_type = raw_response.headers.get("content-type", "")

        if "/json" in media_type:
            try:
                content_data = raw_response.json()
                if content_data is None:
                    # Some HTTP responses, such as a DELETE,
                    # can be successful without any response body.
                    content_data = {}

                parsed_data = ParsedData(status, raw_response, content_data)
            except (TypeError, JSONDecodeError) as error:
                message = f"Failed to process HTTP response : {str(error)}"
                status["content"] = message
                content_data = raw_response.text
                parsed_data = ParsedData(status, raw_response, content_data)

        elif (
            "text/plain" in media_type
            or "text/html" in media_type
            or "text/xml" in media_type
            or "image/" in media_type
        ):
            content_data = raw_response.text
            parsed_data = ParsedData(status, raw_response, content_data)

        else:
            status["content"] = f"Unknown media type returned: {media_type}"
            parsed_data = ParsedData(status, raw_response)

        return parsed_data

    def process_failed_response(self, raw_response: "httpx.Response") -> ParsedData:
        status = {
            "http_status_code": raw_response.status_code,
            "http_reason": get_response_reason(raw_response),
        }
        error_codes = []
        error_messages = []

        try:
            content_data = raw_response.json()

            if not isinstance(content_data, list):
                content_data = [content_data]

            for content_item in content_data:
                content_error = content_item.get("error")

                if content_error:
                    if isinstance(content_error, str):
                        content_error = {"code": None, "message": content_error}

                    status["error"] = content_error
                    error_code = raw_response.status_code
                    error_message = content_error.get("message", "")

                else:
                    error_code = raw_response.status_code
                    error_message = raw_response.text

                if error_code == 403:
                    if not error_message.endswith("."):
                        error_message += ". "

                    error_message += "Contact Refinitiv to check your permissions."

                error_codes.append(error_code)
                error_messages.append(error_message)

        except (TypeError, JSONDecodeError):
            error_codes.append(raw_response.status_code)
            error_messages.append(raw_response.text)

        parsed_data = ParsedData(
            status,
            raw_response,
            error_codes=error_codes,
            error_messages=error_messages,
        )
        return parsed_data

    def parse_raw_response(
        self, raw_response: "httpx.Response"
    ) -> Tuple[bool, ParsedData]:
        is_success = False

        if raw_response is None:
            return is_success, ParsedData({}, {})

        is_success = raw_response.status_code in success_http_codes

        if is_success:
            parsed_data = self.process_successful_response(raw_response)

        else:
            parsed_data = self.process_failed_response(raw_response)

        return is_success, parsed_data



import types
from functools import partial

import pytest

import refinitiv.data as rd
import refinitiv.data.content as rdc
from refinitiv.data.delivery._data._request import Request
from tests.unit.conftest import StubSession, StubResponse
from tests.unit.content.data_for_tests import (
    zc_curves_definitions,
    calendars_definitions,
    universe_definitions,
    fields_in_json_definitions,
    calendars_jsons,
)
from tests.unit.content.test_enums import (
    get_query_params_from_url,
    get_json,
)

defns_with_list_of_items = [
    rdc.custom_instruments.Definition,
    rdc.custom_instruments.events.Definition,
    rdc.custom_instruments.summaries.Definition,
    rdc.esg.basic_overview.Definition,
    rdc.esg.bulk.Definition,
    rdc.esg.full_measures.Definition,
    rdc.esg.full_scores.Definition,
    rdc.esg.standard_measures.Definition,
    rdc.esg.standard_scores.Definition,
    rdc.estimates.view_actuals.annual.Definition,
    rdc.estimates.view_actuals.interim.Definition,
    rdc.estimates.view_actuals_kpi.annual.Definition,
    rdc.estimates.view_actuals_kpi.interim.Definition,
    rdc.estimates.view_summary.annual.Definition,
    rdc.estimates.view_summary.historical_snapshots_non_periodic_measures.Definition,
    rdc.estimates.view_summary.historical_snapshots_periodic_measures_annual.Definition,
    rdc.estimates.view_summary.historical_snapshots_periodic_measures_interim.Definition,
    rdc.estimates.view_summary.historical_snapshots_recommendations.Definition,
    rdc.estimates.view_summary.interim.Definition,
    rdc.estimates.view_summary.non_periodic_measures.Definition,
    rdc.estimates.view_summary.recommendations.Definition,
    rdc.estimates.view_summary_kpi.annual.Definition,
    rdc.estimates.view_summary_kpi.historical_snapshots_kpi.Definition,
    rdc.estimates.view_summary_kpi.interim.Definition,
    rdc.fundamental_and_reference.Definition,
    rdc.historical_pricing.events.Definition,
    rdc.historical_pricing.summaries.Definition,
    rdc.ipa.curves.forward_curves.Definition,
    rdc.ipa.curves.forward_curves.Definitions,
    rdc.ipa.curves.zc_curve_definitions.Definitions,
    rdc.ipa.curves.zc_curves.Definitions,
    rdc.ipa.dates_and_calendars.add_periods.Definition,
    rdc.ipa.dates_and_calendars.add_periods.Definitions,
    rdc.ipa.dates_and_calendars.count_periods.Definition,
    rdc.ipa.dates_and_calendars.count_periods.Definitions,
    rdc.ipa.dates_and_calendars.date_schedule.Definition,
    rdc.ipa.dates_and_calendars.holidays.Definition,
    rdc.ipa.dates_and_calendars.holidays.Definitions,
    rdc.ipa.dates_and_calendars.is_working_day.Definition,
    rdc.ipa.dates_and_calendars.is_working_day.Definitions,
    rdc.ipa.financial_contracts.bond.Definition,
    rdc.ipa.financial_contracts.cap_floor.Definition,
    rdc.ipa.financial_contracts.cds.Definition,
    rdc.ipa.financial_contracts.cross.Definition,
    rdc.ipa.financial_contracts.Definitions,
    rdc.ipa.financial_contracts.option.Definition,
    rdc.ipa.financial_contracts.repo.Definition,
    rdc.ipa.financial_contracts.swap.Definition,
    rdc.ipa.financial_contracts.swaption.Definition,
    rdc.ipa.financial_contracts.term_deposit.Definition,
    rdc.ipa.surfaces.Definitions,
    rdc.ownership.consolidated.breakdown.Definition,
    rdc.ownership.consolidated.concentration.Definition,
    rdc.ownership.consolidated.investors.Definition,
    rdc.ownership.consolidated.recent_activity.Definition,
    rdc.ownership.consolidated.shareholders_report.Definition,
    rdc.ownership.consolidated.top_n_concentration.Definition,
    rdc.ownership.fund.breakdown.Definition,
    rdc.ownership.fund.concentration.Definition,
    rdc.ownership.fund.holdings.Definition,
    rdc.ownership.fund.investors.Definition,
    rdc.ownership.fund.recent_activity.Definition,
    rdc.ownership.fund.shareholders_report.Definition,
    rdc.ownership.fund.top_n_concentration.Definition,
    rdc.ownership.insider.shareholders_report.Definition,
    rdc.ownership.insider.transaction_report.Definition,
    rdc.ownership.investor.holdings.Definition,
    rdc.ownership.org_info.Definition,
    rdc.pricing.Definition,
    rdc.symbol_conversion.Definition,
    rdc.trade_data_service.Definition,
    rd.delivery.cfs.buckets.Definition,
    rd.delivery.omm_stream.Definition,
    rd.delivery.rdp_stream.Definition,
    rd._qpl.fx_swp_to_swp,
]

list_of_items_types = [
    "StrStrings",
    "DefnDefns",
    "OptStrStrs",
    "OptRowHeaders",
    "OptEventTypes",
    "OptAdjustments",
    "OptMarketSession",
    "OptForwardCurveDefinitions",
    "OptHolidayOutputs",
    "OptAssetClass",
    "OptSymbolTypes",
]


def test_count():
    num = len(defns_with_list_of_items)
    assert num == 75, num


def test_definition_has_list_of_items():
    for defn in defns_with_list_of_items:
        if not isinstance(defn, types.FunctionType):
            defn = getattr(defn, "__init__")
        annotations = defn.__annotations__
        if not annotations:
            continue

        has = False
        for param_name, type_notation in annotations.items():
            if type_notation in list_of_items_types:
                has = True
                break

        assert has, defn


def make_checker(param_name, value, func_get_params, response=None):
    def http_request(request, *args, **kwargs):
        nonlocal response
        url = request.url
        if url.startswith(
            "test_get_rdp_url_root/data/pricing/snapshots"
        ) or url.startswith("test_get_rdp_url_root/data/historical-pricing"):
            response = [{}]

        params = func_get_params(request)
        param_value = params.get(param_name)
        assert str(param_value) == str(value), (
            f"Param{param_name}:\n"
            f"Expected value {param_value}. \n"
            f"Value {str(value)}"
        )
        return StubResponse(content_data=response)

    return http_request


def get_request_fundamental_and_reference(request: Request):
    json = request.json
    entity = json.get("Entity", {})
    w = entity.get("W", {})
    requests = w.get("requests", [{}])
    return requests[0]


universe_params = [
    (
        i,
        "universe",
        "universe",
        [
            ("test_universe", "test_universe"),
            (["test_universe"], "test_universe"),
            (["test_universe", "universe"], "test_universe%2Cuniverse"),
        ],
        get_query_params_from_url,
        None,
    )
    for i in universe_definitions
]

fields_in_json = [
    (
        i,
        "fields",
        "fields",
        [
            ("BID", ["BID"]),
            (["BID", "ASK"], ["BID", "ASK"]),
        ],
        get_json,
        None,
    )
    for i in fields_in_json_definitions
]


def get_calendars_json(request: Request):
    json = get_json(request)
    return {"universe": json}


def get_forward_curves(request: Request):
    return get_json(request).get("universe", [{}])[0]


@pytest.mark.parametrize(
    (
        "definition",
        "arg_name",
        "request_param_name",
        "test_data",
        "func_get_params",
        "response",
    ),
    [
        *universe_params,
        (
            partial(rdc.fundamental_and_reference.Definition, fields=[]),
            "universe",
            "instruments",
            [
                ("test_universe", ["TEST_UNIVERSE"]),
                (["test_universe"], ["TEST_UNIVERSE"]),
                (["test_universe", "universe"], ["TEST_UNIVERSE", "UNIVERSE"]),
            ],
            get_request_fundamental_and_reference,
            None,
        ),
        *fields_in_json,
        (
            partial(rdc.fundamental_and_reference.Definition, universe="universe"),
            "fields",
            "fields",
            [
                (["TR.RIC"], [{"name": "TR.RIC"}]),
                (
                    ["TR.RIC", "TR.REVENUE"],
                    [{"name": "TR.RIC"}, {"name": "TR.REVENUE"}],
                ),
            ],
            get_request_fundamental_and_reference,
            None,
        ),
        (
            rdc.ipa.dates_and_calendars.add_periods.Definition,
            "calendars",
            "calendars",
            [
                ("BID", "BID"),
                (["BID", "ASK"], ["BID", "ASK"]),
            ],
            lambda request: get_json(request)[0],
            [{"date": ...}],
        ),
        (
            rdc.ipa.dates_and_calendars.count_periods.Definition,
            "calendars",
            "calendars",
            [
                ("BID", "BID"),
                (["BID", "ASK"], ["BID", "ASK"]),
            ],
            lambda request: get_json(request)[0],
            [{"count": ..., "tenor": ...}],
        ),
        (
            rdc.ipa.dates_and_calendars.date_schedule.Definition,
            "calendars",
            "calendars",
            [
                ("BID", "BID"),
                (["BID", "ASK"], ["BID", "ASK"]),
            ],
            get_json,
            None,
        ),
        (
            rdc.ipa.dates_and_calendars.holidays.Definition,
            "calendars",
            "calendars",
            [
                ("BID", "BID"),
                (["BID", "ASK"], ["BID", "ASK"]),
            ],
            lambda request: get_json(request)[0],
            None,
        ),
        (
            rdc.ipa.dates_and_calendars.is_working_day.Definition,
            "calendars",
            "calendars",
            [
                ("BID", "BID"),
                (["BID", "ASK"], ["BID", "ASK"]),
            ],
            lambda request: get_json(request)[0],
            [{"isWeekEnd": ..., "isWorkingDay": ...}],
        ),
        (
            partial(rdc.historical_pricing.summaries.Definition, universe="universe"),
            "adjustments",
            "adjustments",
            [
                ("unadjusted", "unadjusted"),
                (
                    ["unadjusted", rdc.historical_pricing.Adjustments.CCH],
                    "unadjusted%2CCCH",
                ),
            ],
            get_query_params_from_url,
            [{"universe": {"ric": "universe"}, "data": [{}]}],
        ),
        (
            partial(rdc.historical_pricing.summaries.Definition, universe="universe"),
            "sessions",
            "sessions",
            [
                ("normal", "normal"),
                (["normal", rdc.historical_pricing.MarketSession.PRE], "normal%2Cpre"),
            ],
            get_query_params_from_url,
            [{"universe": {"ric": "universe"}, "data": [{}]}],
        ),
        (
            partial(rdc.pricing.Definition, universe="universe"),
            "fields",
            "fields",
            [
                ("BID", "BID"),
                (["BID", "ASK"], "BID%2CASK"),
            ],
            get_query_params_from_url,
            None,
        ),
        (
            partial(rdc.historical_pricing.events.Definition, universe="universe"),
            "eventTypes",
            "eventTypes",
            [
                ("trade", "trade"),
                (["trade", rdc.historical_pricing.EventTypes.QUOTE], "trade%2Cquote"),
            ],
            get_query_params_from_url,
            [{"universe": {"ric": "universe"}, "data": [{}]}],
        ),
        (
            partial(rdc.historical_pricing.events.Definition, universe="universe"),
            "adjustments",
            "adjustments",
            [
                ("unadjusted", "unadjusted"),
                (
                    ["unadjusted", rdc.historical_pricing.Adjustments.CCH],
                    "unadjusted%2CCCH",
                ),
            ],
            get_query_params_from_url,
            [{"universe": {"ric": "universe"}, "data": [{}]}],
        ),
        (
            rdc.ipa.curves.forward_curves.Definitions,
            "universe",
            "universe",
            [
                (zc_curves_definitions[0], [{"source": "Refinitiv"}]),
                (
                    zc_curves_definitions,
                    [{"source": "Refinitiv"}, {"source": "Peugeot"}],
                ),
            ],
            get_json,
            None,
        ),
        (
            rdc.ipa.curves.zc_curves.Definitions,
            "universe",
            "universe",
            [
                (zc_curves_definitions[0], [{"source": "Refinitiv"}]),
                (
                    zc_curves_definitions,
                    [{"source": "Refinitiv"}, {"source": "Peugeot"}],
                ),
            ],
            get_json,
            None,
        ),
        (
            rdc.ipa.curves.zc_curve_definitions.Definitions,
            "universe",
            "universe",
            [
                (zc_curves_definitions[0], [{"source": "Refinitiv"}]),
                (
                    zc_curves_definitions,
                    [{"source": "Refinitiv"}, {"source": "Peugeot"}],
                ),
            ],
            get_json,
            None,
        ),
        (
            rdc.ipa.dates_and_calendars.add_periods.Definitions,
            "universe",
            "universe",
            [
                (calendars_definitions[0], [calendars_jsons[0]]),
                (calendars_definitions, calendars_jsons),
            ],
            get_calendars_json,
            [{"date": ...}],
        ),
        (
            rdc.ipa.dates_and_calendars.count_periods.Definitions,
            "universe",
            "universe",
            [
                (calendars_definitions[0], [calendars_jsons[0]]),
                (calendars_definitions, calendars_jsons),
            ],
            get_calendars_json,
            [{"count": ..., "tenor": ...}],
        ),
        (
            rdc.ipa.dates_and_calendars.holidays.Definitions,
            "universe",
            "universe",
            [
                (calendars_definitions[0], [calendars_jsons[0]]),
                (calendars_definitions, calendars_jsons),
            ],
            get_calendars_json,
            None,
        ),
        (
            rdc.ipa.dates_and_calendars.is_working_day.Definitions,
            "universe",
            "universe",
            [
                (calendars_definitions[0], [calendars_jsons[0]]),
                (calendars_definitions, calendars_jsons),
            ],
            get_calendars_json,
            [{"isWeekEnd": ..., "isWorkingDay": ...}],
        ),
        (
            rd.delivery.cfs.buckets.Definition,
            "attributes",
            "attributes",
            [
                ("attr1", "attr1"),
                (["attr1", "attr2"], "attr1;attr2"),
            ],
            get_query_params_from_url,
            None,
        ),
        (
            rdc.ipa.curves.forward_curves.Definition,
            "forward_curve_definitions",
            "forwardCurveDefinitions",
            [
                (
                    rdc.ipa.curves.forward_curves.ForwardCurveDefinition(
                        index_tenor="3M",
                        forward_curve_tag="ForwardTag",
                    ),
                    [{"indexTenor": "3M", "forwardCurveTag": "ForwardTag"}],
                )
            ],
            get_forward_curves,
            None,
        ),
        (
            rdc.ipa.dates_and_calendars.add_periods.Definition,
            "currencies",
            "currencies",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
            lambda request: get_json(request)[0],
            [{"date": ...}],
        ),
        (
            rdc.ipa.dates_and_calendars.add_periods.Definition,
            "holiday_outputs",
            "holidayOutputs",
            [
                (["Date"], ["Date"]),
                (rdc.ipa.dates_and_calendars.holidays.HolidayOutputs.DATE, ["Date"]),
                (rdc.ipa.dates_and_calendars.add_periods.HolidayOutputs.DATE, ["Date"]),
                (
                    [
                        "Date",
                        rdc.ipa.dates_and_calendars.add_periods.HolidayOutputs.CALENDARS,
                    ],
                    ["Date", "Calendars"],
                ),
            ],
            lambda request: get_json(request)[0],
            [{"date": ...}],
        ),
        (
            rdc.ipa.dates_and_calendars.count_periods.Definition,
            "currencies",
            "currencies",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
            lambda request: get_json(request)[0],
            [{"count": ..., "tenor": ...}],
        ),
        (
            rdc.ipa.dates_and_calendars.date_schedule.Definition,
            "currencies",
            "currencies",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
            get_json,
            None,
        ),
        (
            rdc.ipa.dates_and_calendars.holidays.Definition,
            "currencies",
            "currencies",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
            lambda request: get_json(request)[0],
            None,
        ),
        (
            rdc.ipa.dates_and_calendars.is_working_day.Definition,
            "currencies",
            "currencies",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
            lambda request: get_json(request)[0],
            [{"isWeekEnd": ..., "isWorkingDay": ...}],
        ),
        (
            rdc.ipa.financial_contracts.Definitions,
            "universe",
            "universe",
            [
                (
                    rdc.ipa.financial_contracts.option.Definition(),
                    [{"instrumentType": "Option", "instrumentDefinition": {}}],
                ),
                (
                    [rdc.ipa.financial_contracts.option.Definition()],
                    [{"instrumentType": "Option", "instrumentDefinition": {}}],
                ),
            ],
            get_json,
            None,
        ),
        (
            rdc.ipa.surfaces.Definitions,
            "universe",
            "universe",
            [
                (rdc.ipa.surfaces.cap.Definition(), [{"underlyingType": "Cap"}]),
                (
                    [
                        rdc.ipa.surfaces.swaption.Definition(),
                        rdc.ipa.surfaces.cap.Definition(),
                    ],
                    [{"underlyingType": "Swaption"}, {"underlyingType": "Cap"}],
                ),
            ],
            get_json,
            None,
        ),
        (
            rdc.symbol_conversion.Definition,
            "symbols",
            "Terms",
            [("symbol1", "symbol1"), (["symbol1", "symbol2"], "symbol1,symbol2")],
            get_json,
            None,
        ),
        (
            partial(rdc.symbol_conversion.Definition, symbols=""),
            "to_symbol_types",
            "ToSymbolTypes",
            [("RIC", None), (["RIC", rdc.symbol_conversion.SymbolTypes.ISIN], None)],
            get_json,
            None,
        ),
        (
            partial(rdc.symbol_conversion.Definition, symbols=""),
            "asset_class",
            "assetClass",
            [
                ("Commodities", None),
                (["Commodities", rdc.symbol_conversion.AssetClass.COMMODITIES], None),
            ],
            get_json,
            None,
        ),
        (
            partial(rdc.historical_pricing.events.Definition, universe="universe"),
            "fields",
            "fields",
            [
                ("BID", "BID%2CDATE_TIME"),
                (["BID", "ASK"], "BID%2CASK%2CDATE_TIME"),
            ],
            get_query_params_from_url,
            None,
        ),
        (
            partial(rdc.historical_pricing.summaries.Definition, universe="universe"),
            "fields",
            "fields",
            [
                ("BID", "BID%2CDATE"),
                (["BID", "ASK"], "BID%2CASK%2CDATE"),
            ],
            get_query_params_from_url,
            None,
        ),
        (
            partial(rd._qpl.fx_swp_to_swp, fx_cross_code=""),
            "tenors",
            "tenors",
            [
                ("1M", ["1M"]),
                (["1M"], ["1M"]),
            ],
            get_json,
            None,
        ),
        (
            partial(rd._qpl.fx_swp_to_swp, fx_cross_code=""),
            "fields",
            "fields",
            [
                ("OutrightCcy1Ccy2Bid", ["OutrightCcy1Ccy2Bid"]),
                (["OutrightCcy1Ccy2Bid"], ["OutrightCcy1Ccy2Bid"]),
            ],
            get_json,
            None,
        ),
    ],
)
def test_list_of_items_get_data_one_request(
    definition, arg_name, request_param_name, test_data, func_get_params, response
):
    session = StubSession(is_open=True)
    rd.session.set_default(session)

    for input_value, expected_value in test_data:
        assert_list_of_items_in_one_request(
            session, request_param_name, expected_value, func_get_params, response
        )
        params = {arg_name: input_value}
        try:
            obj = definition(**params)
            if hasattr(obj, "get_data"):
                obj.get_data(session=session)
        except AttributeError as e:
            if str(e) == "'NoneType' object has no attribute '_get_enum_parameter'":
                continue
            else:
                raise e

    rd.session.set_default(None)


def assert_list_of_items_in_one_request(
    session, request_param_name, test_result, func_get_params, response
):
    session.http_request = make_checker(
        request_param_name, test_result, func_get_params, response
    )


def make_checker_multi_request(expected_values):
    if not isinstance(expected_values, list):
        expected_values = [expected_values]
    expected_values = iter(expected_values)

    def http_request(request, *args, **kwargs):
        url = request.url
        if url.endswith("S%29Instrument.UUID-0000"):
            return StubResponse()

        value = url.rsplit("/", 1)[-1]
        assert value == next(expected_values)

        return StubResponse([{"data": [{}]}])

    return http_request


def make_get_data_bulk(key, value):
    def get_data(*args, **kwargs):
        assert kwargs[key] == value

    return get_data


@pytest.mark.parametrize(
    ("definition", "param_name", "attribute_name", "test_values"),
    [
        (
            rdc.trade_data_service.Definition,
            "universe",
            "_universe",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
        ),
        (
            partial(rdc.trade_data_service.Definition, universe="universe"),
            "fields",
            "_fields",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
        ),
        (
            partial(rdc.trade_data_service.Definition, universe="universe"),
            "filters",
            "_filters",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
        ),
        (
            partial(rd.delivery.omm_stream.Definition, name="universe"),
            "fields",
            "_fields",
            [
                ("val1", ["val1"]),
                (["val1", "val2"], ["val1", "val2"]),
            ],
        ),
        (
            partial(
                rd.delivery.rdp_stream.Definition,
                service="service",
                view="view",
                parameters={},
                api="",
            ),
            "universe",
            "_universe",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
        ),
        (
            partial(
                rd.delivery.rdp_stream.Definition,
                universe="universe",
                service="service",
                parameters={},
                api="",
            ),
            "view",
            "_view",
            [
                ("val1", "val1"),
                (["val1", "val2"], ["val1", "val2"]),
            ],
        ),
        (
            rdc.custom_instruments.Definition,
            "universe",
            "_universe",
            [
                ("val1", ["S)val1."]),
                (["val1", "val2"], ["S)val1.", "S)val2."]),
            ],
        ),
    ],
)
def test_list_of_items_stream(definition, param_name, attribute_name, test_values):
    session = StubSession(is_open=True)
    for test_item in test_values:
        test_value, expected_value = test_item
        stream = definition(**{param_name: test_value}).get_stream(session)
        assert getattr(stream, attribute_name) == expected_value


@pytest.mark.parametrize(
    ("definition", "test_values"),
    [
        (
            rdc.custom_instruments.events.Definition,
            (
                ("universe", "S%29universe."),
                (["universe1", "universe2"], ["S%29universe1.", "S%29universe2."]),
            ),
        ),
        (
            rdc.custom_instruments.summaries.Definition,
            (
                ("universe", "S%29universe."),
                (["universe1", "universe2"], ["S%29universe1.", "S%29universe2."]),
            ),
        ),
        (
            rdc.historical_pricing.events.Definition,
            (
                ("universe", "universe"),
                (["universe1", "universe2"], ["universe1", "universe2"]),
            ),
        ),
        (
            rdc.historical_pricing.summaries.Definition,
            (
                ("universe", "universe"),
                (["universe1", "universe2"], ["universe1", "universe2"]),
            ),
        ),
    ],
)
def test_list_of_items_multi_requests(definition, test_values):
    session = StubSession(is_open=True)
    for test_value in test_values:
        value, expected_result = test_value
        assert_list_of_items_in_multi_request(session, expected_result)
        definition(value).get_data(session=session)


def assert_list_of_items_in_multi_request(session, expected_result):
    session.http_request = make_checker_multi_request(expected_result)


@pytest.mark.parametrize(
    ("params", "param_name", "expected_result"),
    [
        ({"universe": "val1"}, "universe", ["val1"]),
        ({"universe": "universe", "fields": "val1"}, "fields", ["val1"]),
        (
            {"universe": "universe", "fields": ["val1", "val2"]},
            "fields",
            ["val1", "val2"],
        ),
        ({"universe": ["val1", "val2"]}, "universe", ["val1", "val2"]),
    ],
)
def test_bulk_list_of_items(params, param_name, expected_result):
    from refinitiv.data import _configure

    _configure.set_param(
        "bulk.esg.test",
        {
            "db": {
                "connection": {
                    "module": "sqlite3",
                    "parameters": {"database": "file::memory:"},
                }
            },
            "package": {
                "bucket": "test",
                "name": "test",
                "download": {
                    "path": "...",
                },
            },
        },
        True,
    )
    definition = rdc.esg.bulk.Definition("esg.test", **params)
    assert_bulk_list_of_items(definition, param_name, expected_result)
    definition.get_db_data()


def assert_bulk_list_of_items(definition, param_name, expected_result):
    definition._db_manager.get_data = make_get_data_bulk(param_name, expected_result)



import json

import allure
import pytest

import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.errors import RDError
from tests.integration.content.ipa.contracts.swap.conftest import (
    check_http_status_is_success_and_df_value_not_empty,
    check_stream_state_and_df_from_stream,
)
from tests.integration.content.ipa.contracts.swaption.conftest import (
    swaption_definition_01,
    swaption_definition_02,
    swaption_universe,
)
from tests.integration.helpers import (
    get_async_response_from_definition,
    check_dataframe_column_date_for_datetime_type,
)


@allure.suite("Content object - Swaption")
@allure.feature("Content object - Swaption")
@allure.severity(allure.severity_level.CRITICAL)
class TestSwaption:
    @allure.title("Create swaption definition object with valid params and get data")
    @pytest.mark.caseid("36796304")
    def test_get_swaption_analytics(self, open_session):
        response = swaption_definition_01().get_data()
        check_http_status_is_success_and_df_value_not_empty(response)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create swaption definition object with specific session")
    @pytest.mark.caseid("36796339")
    def test_get_swaption_analytics_with_specific_session(self, open_session):
        response = swaption_definition_02().get_data(open_session)
        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title("Create swaption definition object without any parameters")
    @pytest.mark.caseid("36796341")
    def test_get_swaption_analytics_without_any_parameters(self, open_session):
        with pytest.raises(RDError) as error:
            rdf.swaption.Definition().get_data()
        assert (
            str(error.value)
            == "Error code QPS-DPS.1046 | Invalid input: missing underlying instrument. InstrumentDefinition is empty. Please provide a valid InstrumentDefinition."
        )

    @allure.title("Create swaption definition object asynchronously")
    @pytest.mark.caseid("36796342")
    async def test_get_swaption_analytics_asynchronously(self, open_session_async):
        response = await get_async_response_from_definition(swaption_definition_01())
        check_http_status_is_success_and_df_value_not_empty(response)

    @allure.title("Create swaption stream")
    @pytest.mark.caseid("36796343")
    def test_create_swaption_stream(self, open_session):
        stream = swaption_definition_02().get_stream(open_session)
        stream.open()

        check_stream_state_and_df_from_stream(stream)

    @allure.title("Create swaption stream without session")
    @pytest.mark.caseid("36796344")
    def test_create_swaption_stream_without_session(self):
        with pytest.raises(AttributeError) as error:
            swaption_definition_02().get_stream()
        assert (
            str(error.value)
            == "No default session created yet. Please create a session first!"
        )

    @allure.title("Create swaption definition with extended parameters")
    @pytest.mark.caseid("C37474014")
    def test_get_swaption_with_extended_params(self, open_session):
        response = rdf.swaption.Definition(
            instrument_tag="BermudanEURswaption",
            tenor="2Y",
            bermudan_swaption_definition=rdf.swaption.BermudanSwaptionDefinition(
                exercise_schedule_type=rdf.swaption.ExerciseScheduleType.FLOAT_LEG,
                notification_days=22,
            ),
            underlying_definition=rdf.swap.Definition(tenor="1Y", template="NOK_AB6O"),
            extended_params={
                "instrumentDefinition": {
                    "instrumentTag": "BermudanEURswaption",
                    "tenor": "7Y",
                    "notionalAmount": 1000000,
                    "bermudanSwaptionDefinition": {
                        "exerciseScheduleType": "FloatLeg",
                        "notificationDays": 0,
                    },
                    "buySell": "Buy",
                    "exerciseStyle": "BERM",
                    "settlementType": "Cash",
                    "swaptionType": "Payer",
                    "underlyingDefinition": {"tenor": "3Y", "template": "NOK_AB6O"},
                    "strikePercent": 2.75,
                },
                "pricingParameters": {
                    "nbIterations": 80,
                    "valuationDate": "2020-04-24",
                },
            },
        ).get_data()

        check_http_status_is_success_and_df_value_not_empty(response)

        request = json.loads(response.request_message.content.decode("utf-8"))
        request_universe = request.get("universe")[0]
        assert (
            request_universe == swaption_universe
        ), f"Extended params are applied incorrectly"
        assert response.data.df.InstrumentTag[0] == "BermudanEURswaption"



# coding: utf8


from typing import Optional

from ._enums import (
    IndexConvexityAdjustmentIntegrationMethod,
    IndexConvexityAdjustmentMethod,
    PriceSide,
    TenorReferenceDate,
)
from ..._object_definition import ObjectDefinition


class PricingParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    index_convexity_adjustment_integration_method : IndexConvexityAdjustmentIntegrationMethod, optional
        The integration method used for static replication method. the possible values
        are: riemannsum, rungekutta. the default value is 'riemannsum'.
    index_convexity_adjustment_method : IndexConvexityAdjustmentMethod, optional
        The convexity adjustment type for constant maturity swaps (cms) and libor
        in-arrears swaps. the possible values are: none, blackscholes, linearswapmodel,
        replication. the default value is 'blackscholes'.
    price_side : PriceSide, optional
        The quoted price side of the instrument. optional. default value is 'mid'.
    tenor_reference_date : TenorReferenceDate, optional
        In case the swap definition uses 'starttenor', 'starttenor' defines whether the
        swap start date is calculated from valuation date or from spot date
    discounting_ccy : str, optional
        The currency code, which defines the choice of the discounting curve. the value
        is expressed in iso 4217 alphabetical format (e.g. 'usd'). by default,
        settlementccy or the paid leg currency is used.
    discounting_tenor : str, optional

    market_data_date : str, optional
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
    market_value_in_deal_ccy : float, optional
        The dirty market value of the instrument computed as
        [cleanmarketvalueindealccy+accruedamountindealccy]. the value is expressed in
        the deal currency. the default value is '0'.
    report_ccy : str, optional
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
    use_legs_signing : bool, optional
        Enable the signing of "risk measures" and "valuation" outputs based on leg's
        direction the default value is false.
    valuation_date : str, optional
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
    """

    def __init__(
        self,
        index_convexity_adjustment_integration_method: Optional[
            IndexConvexityAdjustmentIntegrationMethod
        ] = None,
        index_convexity_adjustment_method: Optional[
            IndexConvexityAdjustmentMethod
        ] = None,
        price_side: Optional[PriceSide] = None,
        tenor_reference_date: Optional[TenorReferenceDate] = None,
        discounting_ccy: Optional[str] = None,
        discounting_tenor: Optional[str] = None,
        market_data_date: Optional[str] = None,
        market_value_in_deal_ccy: Optional[float] = None,
        report_ccy: Optional[str] = None,
        use_legs_signing: Optional[bool] = None,
        valuation_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.index_convexity_adjustment_integration_method = (
            index_convexity_adjustment_integration_method
        )
        self.index_convexity_adjustment_method = index_convexity_adjustment_method
        self.price_side = price_side
        self.tenor_reference_date = tenor_reference_date
        self.discounting_ccy = discounting_ccy
        self.discounting_tenor = discounting_tenor
        self.market_data_date = market_data_date
        self.market_value_in_deal_ccy = market_value_in_deal_ccy
        self.report_ccy = report_ccy
        self.use_legs_signing = use_legs_signing
        self.valuation_date = valuation_date

    @property
    def index_convexity_adjustment_integration_method(self):
        """
        The integration method used for static replication method. the possible values
        are: riemannsum, rungekutta. the default value is 'riemannsum'.
        :return: enum IndexConvexityAdjustmentIntegrationMethod
        """
        return self._get_enum_parameter(
            IndexConvexityAdjustmentIntegrationMethod,
            "indexConvexityAdjustmentIntegrationMethod",
        )

    @index_convexity_adjustment_integration_method.setter
    def index_convexity_adjustment_integration_method(self, value):
        self._set_enum_parameter(
            IndexConvexityAdjustmentIntegrationMethod,
            "indexConvexityAdjustmentIntegrationMethod",
            value,
        )

    @property
    def index_convexity_adjustment_method(self):
        """
        The convexity adjustment type for constant maturity swaps (cms) and libor
        in-arrears swaps. the possible values are: none, blackscholes, linearswapmodel,
        replication. the default value is 'blackscholes'.
        :return: enum IndexConvexityAdjustmentMethod
        """
        return self._get_enum_parameter(
            IndexConvexityAdjustmentMethod, "indexConvexityAdjustmentMethod"
        )

    @index_convexity_adjustment_method.setter
    def index_convexity_adjustment_method(self, value):
        self._set_enum_parameter(
            IndexConvexityAdjustmentMethod, "indexConvexityAdjustmentMethod", value
        )

    @property
    def price_side(self):
        """
        The quoted price side of the instrument. optional. default value is 'mid'.
        :return: enum PriceSide
        """
        return self._get_enum_parameter(PriceSide, "priceSide")

    @price_side.setter
    def price_side(self, value):
        self._set_enum_parameter(PriceSide, "priceSide", value)

    @property
    def tenor_reference_date(self):
        """
        In case the swap definition uses 'starttenor', 'starttenor' defines whether the
        swap start date is calculated from valuation date or from spot date
        :return: enum TenorReferenceDate
        """
        return self._get_enum_parameter(TenorReferenceDate, "tenorReferenceDate")

    @tenor_reference_date.setter
    def tenor_reference_date(self, value):
        self._set_enum_parameter(TenorReferenceDate, "tenorReferenceDate", value)

    @property
    def discounting_ccy(self):
        """
        The currency code, which defines the choice of the discounting curve. the value
        is expressed in iso 4217 alphabetical format (e.g. 'usd'). by default,
        settlementccy or the paid leg currency is used.
        :return: str
        """
        return self._get_parameter("discountingCcy")

    @discounting_ccy.setter
    def discounting_ccy(self, value):
        self._set_parameter("discountingCcy", value)

    @property
    def discounting_tenor(self):
        """
        :return: str
        """
        return self._get_parameter("discountingTenor")

    @discounting_tenor.setter
    def discounting_tenor(self, value):
        self._set_parameter("discountingTenor", value)

    @property
    def market_data_date(self):
        """
        The date at which the market data is retrieved. the value is expressed in iso
        8601 format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). it
        should be less or equal tovaluationdate). optional. by
        default,marketdatadateisvaluationdateor today.
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def market_value_in_deal_ccy(self):
        """
        The dirty market value of the instrument computed as
        [cleanmarketvalueindealccy+accruedamountindealccy]. the value is expressed in
        the deal currency. the default value is '0'.
        :return: float
        """
        return self._get_parameter("marketValueInDealCcy")

    @market_value_in_deal_ccy.setter
    def market_value_in_deal_ccy(self, value):
        self._set_parameter("marketValueInDealCcy", value)

    @property
    def report_ccy(self):
        """
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def use_legs_signing(self):
        """
        Enable the signing of "risk measures" and "valuation" outputs based on leg's
        direction the default value is false.
        :return: bool
        """
        return self._get_parameter("useLegsSigning")

    @use_legs_signing.setter
    def use_legs_signing(self, value):
        self._set_parameter("useLegsSigning", value)

    @property
    def valuation_date(self):
        """
        The date at which the instrument is valued. the value is expressed in iso 8601
        format: yyyy-mm-ddt[hh]:[mm]:[ss]z (e.g., '2021-01-01t00:00:00z'). by default,
        marketdatadate is used. if marketdatadate is not specified, the default value is
        today.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)
